import unittest

import numpy as np

from Orange.data.table import Table, Domain, ContinuousVariable

from orangecontrib.bioinformatics.preprocess import ZScore, LogarithmicScale, QuantileNormalization


class TestQuantileNormalization(unittest.TestCase):
    def test_quantile_normalization(self):
        domain = Domain(
            [ContinuousVariable('A'), ContinuousVariable('B'), ContinuousVariable('C'), ContinuousVariable('D')]
        )
        table = Table.from_list(domain, [[5, 2, 3, 4], [4, 1, 4, 2], [3, 4, 6, 8]])
        _table = QuantileNormalization()(table)

        # expected result
        result = np.array([[5.66666667, 2, 3, 4.66666667], [5.166667, 2, 5.166667, 3], [2, 3, 4.66666667, 5.66666667]])
        np.testing.assert_array_almost_equal(result, _table.X)


class TestZScore(unittest.TestCase):
    def test_z_score(self):
        domain = Domain([ContinuousVariable('A'), ContinuousVariable('B')])
        table = Table.from_list(domain, [[1, 2], [3, 4]])

        _table = ZScore(axis=0)(table)
        np.testing.assert_array_almost_equal([[-1, -1], [1, 1]], _table.X)

        _table = ZScore(axis=1)(table)
        np.testing.assert_array_almost_equal([[-1, 1], [-1, 1]], _table.X)


class TestLogarithmicScale(unittest.TestCase):
    def test_log2_normalization(self):
        domain = Domain([ContinuousVariable('A'), ContinuousVariable('B')])
        table = Table.from_list(domain, [[0, 1], [3, 7], [15, 31]])

        _table = LogarithmicScale()(table)
        np.testing.assert_array_almost_equal([[0, 1], [2, 3], [4, 5]], _table.X)


if __name__ == '__main__':
    unittest.main()



import os
from unittest.mock import Mock

from orangewidget.tests.base import WidgetTest

from Orange.data import Table, Domain, StringVariable

from orangecontrib.bioinformatics.widgets.OWGeneSets import OWGeneSets
from orangecontrib.bioinformatics.widgets.utils.data import TableAnnotation


class TestOWGeneSets(WidgetTest):
    def setUp(self) -> None:
        self._path = os.path.join(os.path.dirname(__file__), 'data')
        self.widget = self.create_widget(OWGeneSets)

        domain = Domain([], metas=[StringVariable('name'), StringVariable('Entrez ID')])
        self.data = Table.from_list(domain, [['CA1', '759'], ['CA2', '760'], ['CA3', '761']])
        self.data.attributes[TableAnnotation.tax_id] = '9606'
        self.data.attributes[TableAnnotation.gene_as_attr_name] = False
        self.data.attributes[TableAnnotation.gene_id_column] = 'Entrez ID'

    def test_input_info(self):
        input_sum = self.widget.info.set_input_summary = Mock()

        self.send_signal(self.widget.Inputs.data, self.data)
        self.wait_until_finished()
        input_sum.assert_called_with("3", "3 unique gene names on input.\n")

        self.send_signal(self.widget.Inputs.custom_gene_sets, self.data)
        self.wait_until_finished()
        input_sum.assert_called_with("3|3", "3 unique gene names on input.\n3 marker genes in 3 sets\n")

        self.send_signal(self.widget.Inputs.data, None)
        self.send_signal(self.widget.Inputs.custom_gene_sets, None)
        self.wait_until_finished()
        input_sum.assert_called_with(self.widget.info.NoInput)

    def test_output_info(self):
        output_sum = self.widget.info.set_output_summary = Mock()

        self.send_signal(self.widget.Inputs.data, self.data)
        self.wait_until_finished()
        output_sum.assert_called_with("0", "0 genes on output.\n")

        self.send_signal(self.widget.Inputs.data, None)
        output_sum.assert_called_with(self.widget.info.NoOutput)



""" KEGG api interface. """
from __future__ import absolute_import

import os
import warnings
from datetime import datetime
from operator import itemgetter
from contextlib import closing

import six

from orangecontrib.bioinformatics.kegg import caching
from orangecontrib.bioinformatics.kegg.types import Link, BInfo, Definition, OrganismSummary
from orangecontrib.bioinformatics.kegg.caching import touch_dir, cache_entry, cached_method
from orangecontrib.bioinformatics.kegg.service import web_service

# A list of all databases with names, abbreviations
DATABASES = [
    ("KEGG Pathway", "pathway", "path", None),
    ("KEGG Brite", "brite", "br", None),
    ("KEGG Module", "module", "md", "M"),
    ("KEGG Disease", "disease", "ds", "H"),
    ("KEGG Drug", "drug", "dr", "D"),
    ("KEGG Orthology", "orthology", "ko", "K"),
    ("KEGG Genome", "genome", "genome", "T"),
    ("KEGG Genomes", "genomes", "gn", "T"),
    ("KEGG Genes", "genes", None, None),
    ("KEGG Ligand", "ligand", "ligand", None),
    ("KEGG Compound", "compound", "cpd", "C"),
    ("KEGG Glycan", "glycan", "gl", "G"),
    ("KEGG Reaction", "reaction", "rn", "R"),
    ("KEGG RPair", "rpair", "rp", "RP"),
    ("KEGG RClass", "rclass", "rc", "RC"),
    ("KEGG Enzyme", "enzyme", "ec", "E"),
]


def _link_targets(links):
    return sorted(set(map(itemgetter(1), links)))


class KeggApi(object):
    """
    An abstraction of a rest KEGG API.
    """

    def __init__(self):
        self.service = web_service()

    def list_organisms(self):
        """
        Return a list of all available organisms,

        >>> api.list_organisms()  # doctest: +ELLIPSIS
        [OrganismSummary(entry_id='T01001', ...
        """
        return list(map(OrganismSummary.from_str, self.service.list.organism.get().splitlines()))

    def list_pathways(self, organism):
        """
        Return a list of all available pathways for `organism`

        >>> api.list_pathways("hsa")  # doctest: +ELLIPSIS
        [Definition(entry_id='path:hsa00010', ...
        """
        return list(map(Definition.from_str, self.service.list.pathway(organism).get().splitlines()))

    def list(self, db):
        """
        Return a list of all available entries in database `db`.
        """
        return list(map(Definition.from_str, self.service.list(db).get().splitlines()))

    #######
    # DBGET
    #######

    def info(self, db):
        """
        Return info for database `db`

        >>> print(api.info("pathway"))
        BInfo(entry_id='path', definition='KEGG Pathway Database', ...
        """
        result = self.service.info(db).get()
        return BInfo.from_text(result)

    def find(self, db, keywords):
        """
        Search database 'db' for keywords.
        """
        if isinstance(keywords, six.string_types):
            keywords = [keywords]

        return self.service.find(db)("+".join(keywords)).get()

    def get(self, ids):
        """
        Retrieve database entries for `ids` list.
        """
        if not isinstance(ids, six.string_types):
            # Sequence of ids
            ids = "+".join(ids)

        return self.service.get(ids).get()

    def conv(self, target_db, source):
        """
        Return a mapping from source to target_db ids as a list of two
        tuples [(source_id, target_id), ...].

        """
        if not isinstance(source, six.string_types):
            source = "+".join(source)

        res = self.service.conv(target_db)(source).get()
        return [tuple(line.split("\t")) for line in res.splitlines()]

    def link(self, target_db, source_db=None, ids=None):
        if not (source_db or ids):
            raise ValueError("One of 'source_db' or 'ids' must be supplied")
        if source_db and ids:
            raise ValueError("Only one 'source_db' or 'ids' must be supplied")

        if source_db:
            result = self.service.link(target_db)(source_db).get()
        else:
            result = self.service.link(target_db)("+".join(ids)).get()

        return list(map(Link._make, map(str.split, result.splitlines())))

    def get_genes_by_enzyme(self, enzyme_id, org):
        return _link_targets(self.link(org, ids=[enzyme_id]))

    def get_enzymes_by_gene(self, gene_id):
        return _link_targets(self.link("ec", ids=[gene_id]))

    def get_enzymes_by_compound(self, compound_id):
        return _link_targets(self.link("ec", ids=[compound_id]))

    def get_enzymes_by_glycan(self, glycan_id):
        return _link_targets(self.link("ec", ids=[glycan_id]))

    def get_enzymes_by_reaction(self, reaction_id):
        return _link_targets(self.link("ec", ids=[reaction_id]))

    def get_compounds_by_enzyme(self, enzyme_id):
        return _link_targets(self.link("compound", ids=[enzyme_id]))

    def get_compounds_by_reaction(self, reaction_id):
        return _link_targets(self.link("compound", ids=[reaction_id]))

    def get_glycans_by_enzyme(self, enzyme_id):
        return _link_targets(self.link("gl", ids=[enzyme_id]))

    def get_glycans_by_reaction(self, reaction_id):
        return _link_targets(self.link("gl", ids=[reaction_id]))

    def get_reactions_by_enzyme(self, enzyme_id):
        return _link_targets(self.link("rn", ids=[enzyme_id]))

    def get_reactions_by_compound(self, compound_id):
        return _link_targets(self.link("rn", ids=[compound_id]))

    def get_reactions_by_glycan(self, glycan_id):
        return _link_targets(self.link("rn", ids=[glycan_id]))

    ######
    # SSDB
    ######

    # No replacement api in the KEGG REST api.
    def get_best_best_neighbors_by_gene(self, genes_id, offset, limit):
        raise NotImplementedError

    def get_best_neighbors_by_gene(self, genes_id, offset, limit):
        raise NotImplementedError

    def get_reverse_best_neighbors_by_gene(self, genes_id, offset, limit):
        raise NotImplementedError

    def get_paralogs_by_gene(self, genes_id, offset, limit):
        raise NotImplementedError

    #######
    # Motif
    #######

    # No replacement api in KEGG REST api
    def get_motifs_by_gene(self, genes_id, db):
        raise NotImplementedError

    def get_genes_by_motifs(self, motif_id_list, offset, limit):
        raise NotImplementedError

    ####
    # KO
    ####

    def get_ko_by_gene(self, genes_id):
        raise NotImplementedError

    def get_ko_by_ko_class(self, ko_class_id):
        raise NotImplementedError

    def get_genes_by_ko_class(self, ko_class_id, org, offset, limit):
        raise NotImplementedError

    def get_genes_by_ko(self, ko_id, org):
        raise NotImplementedError

    #########
    # Pathway
    #########

    def mark_pathway_by_objects(self, pathway_id, object_id_list):
        raise NotImplementedError

    def color_pathway_by_objects(self, pathway_id, object_id_list, fg_color_list, bg_color_list):
        raise NotImplementedError

    def color_pathway_by_elements(self, pathway_id, element_id_list, fg_color_list, bg_color_list):
        raise NotImplementedError

    def get_html_of_marked_pathway_by_objects(self, pathway_id, object_id_list):
        raise NotImplementedError

    def get_html_of_colored_pathway_by_objects(self, pathway_id, object_id_list, fg_color_list, bg_color_list):
        raise NotImplementedError

    def get_html_of_colored_pathway_by_elements(self, pathway_id, element_id_list, fg_color_list, bg_color_list):
        raise NotImplementedError

    def get_references_by_pathway(self, pathway_id):
        return self.service.get_references_by_pathway(pathway_id)

    def get_element_relations_by_pathway(self, pathway_id):
        return self.service.get_element_relations_by_pathway(pathway_id)

    def get_genes_by_organism(self, organism, offset=None, limit=None):
        if offset is not None:
            raise NotImplementedError("offset is no longer supported")
        if limit is not None:
            raise NotImplementedError("limit is no longer supported.")

        res = self.service.list(organism).get().splitlines()
        return [r.split(None, 1)[0] for r in res]

    def get_number_of_genes_by_organism(self, organism):
        raise NotImplementedError

    ####################
    # Objects by pathway
    ####################

    def get_elements_by_pathway(self, pathway_id):
        raise NotImplementedError

    def get_genes_by_pathway(self, pathway_id):
        return _link_targets(self.link("genes", ids=[pathway_id]))

    def get_enzymes_by_pathway(self, pathway_id):
        return _link_targets(self.link("ec", ids=[pathway_id]))

    def get_compounds_by_pathway(self, pathway_id):
        return _link_targets(self.link("compound", ids=[pathway_id]))

    def get_drugs_by_pathway(self, pathway_id):
        return _link_targets(self.link("drug", ids=[pathway_id]))

    def get_glycans_by_pathway(self, pathway_id):
        return _link_targets(self.link("gl", ids=[pathway_id]))

    def get_reactions_by_pathway(self, pathway_id):
        return _link_targets(self.link("rn", ids=[pathway_id]))

    def get_kos_by_pathway(self, pathway_id):
        return _link_targets(self.link("ko", ids=[pathway_id]))

    #############################################
    # Pathways and genes of a specific organism #
    #############################################

    def get_genes_pathway_organism(self, organism):
        l = self.link("pathway", organism)
        return list(map(tuple, l))

    #####################
    # Pathways by objects
    #####################

    # These functions returned results intersections.
    def get_pathways_by_genes(self, gene_list):
        raise NotImplementedError

    def get_pathways_by_enzymes(self, enzyme_list):
        raise NotImplementedError

    def get_pathways_by_compounds(self, compound_list):
        raise NotImplementedError

    def get_pathways_by_drugs(self, drug_list):
        raise NotImplementedError

    def get_pathways_by_glycans(self, glycan_list):
        raise NotImplementedError

    def get_pathways_by_reactions(self, reaction_list):
        raise NotImplementedError

    def get_pathways_by_kos(self, ko_list):
        raise NotImplementedError

    ##########################
    # Relations among pathways
    ##########################

    def get_linked_pathways(self, pathway_id):
        if not pathway_id.startswith("path:"):
            pathway_id = "path:" + pathway_id
        return _link_targets(self.link("pathway", ids=[pathway_id]))


"""
KEGG api with caching
"""


try:
    from functools import lru_cache
except ImportError:
    # TODO: move a copy of lru_cache in .caching if distributing this as a
    # standalone package
    from Orange.utils import lru_cache


class CachedKeggApi(KeggApi):
    def __init__(self, store=None):
        KeggApi.__init__(self)
        if store is None:
            self.store = {}

    # Needed API for cached decorator.
    def cache_store(self):
        from . import conf

        path = conf.params["cache.path"]
        touch_dir(path)
        return caching.Sqlite3Store(os.path.join(path, "kegg_api_cache_2.sqlite3"))

    def last_modified(self, args, kwargs=None):
        return getattr(self, "default_release", "")

    def set_default_release(self, release):
        self.default_release = release

    @cached_method
    def list_organisms(self):
        return KeggApi.list_organisms(self)

    @cached_method
    def list_pathways(self, organism):
        return KeggApi.list_pathways(self, organism)

    @cached_method
    def list(self, db):
        return KeggApi.list(self, db)

    @lru_cache()  # not persistently cached
    def info(self, db):
        return KeggApi.info(self, db)

    @cached_method
    def find(self, db, keywords):
        return KeggApi.find(self, db, keywords)

    @cached_method
    def get(self, ids):
        if not isinstance(ids, six.string_types):
            return self._batch_get(ids)
        else:
            return KeggApi.get(self, ids)

    @cached_method
    def link(self, target_db, source_db=None, ids=None):
        return KeggApi.link(self, target_db, source_db, ids)

    def _batch_get(self, ids):
        if len(ids) > 10:
            raise ValueError("Can batch at most 10 ids at a time.")

        get = self.get
        uncached = []
        unmatched = set()

        with closing(get.cache_store()) as store:
            # Which ids are already cached
            # TODO: Invalidate entries by release string.
            for id in ids:
                key = get.key_from_args((id,))
                if not get.key_has_valid_cache(key, store):
                    uncached.append(id)

        if uncached:
            # in case there are duplicate ids
            uncached = sorted(set(uncached))

            rval = KeggApi.get(self, uncached)

            if rval is not None:
                entries = rval.split("///\n")
            else:
                entries = []

            if entries and not entries[-1].strip():
                # Delete the last single newline entry if present
                del entries[-1]

            if len(entries) != len(uncached):
                new_uncached, entries = match_by_ids(uncached, entries)
                unmatched = set(uncached) - set(new_uncached)
                uncached = new_uncached
                warnings.warn("Unable to match entries for keys: %s." % ", ".join(map(repr, unmatched)))

            with closing(get.cache_store()) as store:
                for id, entry in zip(uncached, entries):
                    key = get.key_from_args((id,))
                    if entry is not None:
                        entry = entry + "///\n"
                    store[key] = cache_entry(entry, mtime=datetime.now())

        # Finally join all the results, but drop all None objects

        with closing(get.cache_store()):
            keys = [get.key_from_args((id,)) for id in ids]
            entries = [store[key].value for key in keys]

        entries = filter(lambda e: e is not None, entries)

        rval = "".join(entries)
        return rval

    @cached_method
    def conv(self, target_db, source):
        return KeggApi.conv(self, target_db, source)

    ########
    # LinkDB
    ########

    @cached_method
    def get_genes_by_enzyme(self, enzyme_id, org):
        return KeggApi.get_genes_by_enzyme(self, enzyme_id, org)

    @cached_method
    def get_enzymes_by_gene(self, genes_id):
        return KeggApi.get_enzymes_by_gene(self, genes_id)

    @cached_method
    def get_enzymes_by_compound(self, compound_id):
        return KeggApi.get_enzymes_by_compound(self, compound_id)

    @cached_method
    def get_enzymes_by_glycan(self, glycan_id):
        return KeggApi.get_enzymes_by_glycan(self, glycan_id)

    @cached_method
    def get_enzymes_by_reaction(self, reaction_id):
        return KeggApi.get_enzymes_by_reaction(self, reaction_id)

    @cached_method
    def get_compounds_by_enzyme(self, enzyme_id):
        return KeggApi.get_compounds_by_enzyme(self, enzyme_id)

    @cached_method
    def get_compounds_by_reaction(self, reaction_id):
        return KeggApi.get_compounds_by_reaction(self, reaction_id)

    @cached_method
    def get_glycans_by_enzyme(self, enzyme_id):
        return KeggApi.get_glycans_by_enzyme(self, enzyme_id)

    @cached_method
    def get_glycans_by_reaction(self, reaction_id):
        return KeggApi.get_glycans_by_reaction(self, reaction_id)

    @cached_method
    def get_reactions_by_enzyme(self, enzyme_id):
        return KeggApi.get_reactions_by_enzyme(self, enzyme_id)

    @cached_method
    def get_reactions_by_compound(self, compound_id):
        return KeggApi.get_reactions_by_compound(self, compound_id)

    @cached_method
    def get_reactions_by_glycan(self, glycan_id):
        return KeggApi.get_reactions_by_glycan(self, glycan_id)

    ######
    # SSDB
    ######

    @cached_method
    def get_best_best_neighbors_by_gene(self, genes_id, offset, limit):
        return KeggApi.get_best_best_neighbors_by_gene(self, genes_id, offset, limit)

    @cached_method
    def get_best_neighbors_by_gene(self, genes_id, offset, limit):
        return KeggApi.get_best_neighbors_by_gene(self, genes_id, offset, limit)

    @cached_method
    def get_reverse_best_neighbors_by_gene(self, genes_id, offset, limit):
        return KeggApi.get_reverse_best_neighbors_by_gene(self, genes_id, offset, limit)

    @cached_method
    def get_paralogs_by_gene(self, genes_id, offset, limit):
        return KeggApi.get_paralogs_by_gene(self, genes_id, offset, limit)

    #######
    # Motif
    #######

    @cached_method
    def get_motifs_by_gene(self, genes_id, db):
        return KeggApi.get_motifs_by_gene(self, genes_id, db)

    @cached_method
    def get_genes_by_motifs(self, motif_id_list, offset, limit):
        return KeggApi.get_genes_by_motifs(self, motif_id_list, offset, limit)

    ####
    # KO
    ####

    @cached_method
    def get_ko_by_gene(self, genes_id):
        return KeggApi.get_ko_by_gene(self, genes_id)

    @cached_method
    def get_ko_by_ko_class(self, ko_class_id):
        return KeggApi.service.get_ko_by_ko_class(self, ko_class_id)

    @cached_method
    def get_genes_by_ko_class(self, ko_class_id, org, offset, limit):
        return KeggApi.get_genes_by_ko_class(self, ko_class_id, org, offset, limit)

    @cached_method
    def get_genes_by_ko(self, ko_id, org):
        return KeggApi.get_genes_by_ko(self, ko_id, org)

    #########
    # Pathway
    #########

    @cached_method
    def get_genes_by_organism(self, organism, offset=None, limit=None):
        return KeggApi.get_genes_by_organism(self, organism, offset=offset, limit=limit)

    @cached_method
    def get_number_of_genes_by_organism(self, organism):
        return KeggApi.get_number_of_genes_by_organism(self, organism)

    @cached_method
    def get_pathways_by_genes(self, gene_list):
        return KeggApi.get_pathways_by_genes(self, gene_list)

    @cached_method
    def get_pathways_by_enzymes(self, enzyme_list):
        return KeggApi.get_pathways_by_enzymes(self, enzyme_list)

    @cached_method
    def get_pathways_by_compounds(self, compound_list):
        return KeggApi.get_pathways_by_compounds(self, compound_list)

    @cached_method
    def get_pathways_by_drugs(self, drug_list):
        return KeggApi.get_pathways_by_drugs(self, drug_list)

    @cached_method
    def get_pathways_by_glycans(self, glycan_list):
        return KeggApi.get_pathways_by_glycans(self, glycan_list)

    @cached_method
    def get_pathways_by_reactions(self, reaction_list):
        return KeggApi.get_pathways_by_reactions(self, reaction_list)

    @cached_method
    def get_pathways_by_kos(self, ko_list):
        return KeggApi.get_pathways_by_kos(self, ko_list)

    @cached_method
    def get_elements_by_pathway(self, pathway_id):
        return KeggApi.get_elements_by_pathway(self, pathway_id)

    @cached_method
    def get_genes_by_pathway(self, pathway_id):
        return KeggApi.get_genes_by_pathway(self, pathway_id)

    @cached_method
    def get_enzymes_by_pathway(self, pathway_id):
        return KeggApi.get_enzymes_by_pathway(self, pathway_id)

    @cached_method
    def get_compounds_by_pathway(self, pathway_id):
        return KeggApi.get_compounds_by_pathway(self, pathway_id)

    @cached_method
    def get_drugs_by_pathway(self, pathway_id):
        return KeggApi.get_drugs_by_pathway(self, pathway_id)

    @cached_method
    def get_glycans_by_pathway(self, pathway_id):
        return KeggApi.get_glycans_by_pathway(self, pathway_id)

    @cached_method
    def get_reactions_by_pathway(self, pathway_id):
        return KeggApi.get_reactions_by_pathway(self, pathway_id)

    @cached_method
    def get_kos_by_pathway(self, pathway_id):
        return KeggApi.get_kos_by_pathway(self, pathway_id)

    @cached_method
    def get_genes_pathway_organism(self, org):
        return KeggApi.get_genes_pathway_organism(self, org)


def match_by_ids(ids, entries):
    """

    """

    unmatched_ids = set(ids)
    unmatched_entries = set(entries)

    matched_ids = []
    matched_entries = []

    def match_add(search_id, entry):
        """
        Move search_id and entry to the matched lists.
        """
        matched_ids.append(search_id)
        matched_entries.append(entry)

        # Remove from the unmatched set
        unmatched_ids.remove(search_id)
        unmatched_entries.remove(entry)

    def entry_split(entry_text):
        line, _ = entry_text.split("\n", 1)
        return line.split(None, 2)

    entries_by_id = {}

    for entry in entries:
        _, eid, _ = entry_split(entry)
        entries_by_id[eid] = entry

    # First match full search ids
    for search_id in list(unmatched_ids):
        if search_id in entries_by_id:
            entry = entries_by_id.pop(search_id)
            match_add(search_id, entry)

    # Second pass, split the search ids by ':' to db and identifier part,
    # match by identifier
    for search_id in list(unmatched_ids):
        if ":" in search_id:
            db_id, rest = search_id.split(":", 1)
            if rest in entries_by_id:
                entry = entries_by_id.pop(rest)
                match_add(search_id, entry)

    return matched_ids, matched_entries



""" GeneSets """
from enum import IntEnum
from types import SimpleNamespace
from typing import Set, List, Tuple, Optional
from urllib.parse import urlparse

from AnyQt.QtGui import QColor, QStandardItem, QStandardItemModel
from AnyQt.QtCore import Qt, QSize
from AnyQt.QtWidgets import QTreeView, QHBoxLayout, QHeaderView

from Orange.data import Table, Domain
from Orange.data import filter as table_filter
from Orange.widgets.gui import LinkRole, LinkStyledItemDelegate, spin, vBox, lineEdit, widgetBox, auto_commit
from Orange.widgets.widget import Msg, OWWidget
from Orange.widgets.settings import Setting, SettingProvider
from Orange.widgets.utils.signals import Input, Output
from Orange.widgets.utils.concurrent import TaskState, ConcurrentWidgetMixin

from orangecontrib.bioinformatics.geneset import GeneSets
from orangecontrib.bioinformatics.widgets.utils.gui import FilterProxyModel, NumericalColumnDelegate
from orangecontrib.bioinformatics.widgets.components import GeneSetSelection
from orangecontrib.bioinformatics.widgets.utils.data import TableAnnotation, check_table_annotation


class Results(SimpleNamespace):
    items: List[QStandardItem] = []


def run(gene_sets: GeneSets, selected_gene_sets: List[Tuple[str, ...]], genes, state: TaskState) -> Results:
    results = Results()
    items = []
    step, steps = 0, len(gene_sets)

    if not genes:
        return results

    state.set_status('Calculating...')

    for gene_set in sorted(gene_sets):

        step += 1
        if step % (steps / 10) == 0:
            state.set_progress_value(100 * step / steps)

        if gene_set.hierarchy not in selected_gene_sets:
            continue

        if state.is_interruption_requested():
            return results

        matched_set = gene_set.genes & genes
        if len(matched_set) > 0:
            category_column = QStandardItem()
            term_column = QStandardItem()
            count_column = QStandardItem()
            genes_column = QStandardItem()

            category_column.setData(", ".join(gene_set.hierarchy), Qt.DisplayRole)
            term_column.setData(gene_set.name, Qt.DisplayRole)
            term_column.setData(gene_set.name, Qt.ToolTipRole)

            # there was some cases when link string was not empty string but not valid (e.g. "_")
            if gene_set.link and urlparse(gene_set.link).scheme:
                term_column.setData(gene_set.link, LinkRole)
                term_column.setForeground(QColor(Qt.blue))

            count_column.setData(matched_set, Qt.UserRole)
            count_column.setData(len(matched_set), Qt.DisplayRole)

            genes_column.setData(len(gene_set.genes), Qt.DisplayRole)
            genes_column.setData(set(gene_set.genes), Qt.UserRole)  # store genes to get then on output on selection

            items.append([count_column, genes_column, category_column, term_column])

    results.items = items
    return results


class Header(IntEnum):
    count = 0
    genes = 1
    category = 2
    term = 3

    @staticmethod
    def labels():
        return ['Count', 'Genes In Set', 'Category', 'Term']


class OWGeneSets(OWWidget, ConcurrentWidgetMixin):
    name = 'Gene Sets'
    description = ""
    icon = 'icons/OWGeneSets.svg'
    priority = 80
    want_main_area = True

    organism = Setting(None, schema_only=True)
    stored_gene_sets_selection = Setting([], schema_only=True)
    selected_rows = Setting([], schema_only=True)

    min_count: int
    min_count = Setting(5)

    use_min_count: bool
    use_min_count = Setting(True)

    auto_commit: bool
    auto_commit = Setting(False)

    search_pattern: str
    search_pattern = Setting('')

    # component settings
    gs_selection_component: SettingProvider = SettingProvider(GeneSetSelection)

    class Inputs:
        data = Input('Data', Table)
        custom_gene_sets = Input('Custom Gene Sets', Table)

    class Outputs:
        matched_genes = Output('Matched Genes', Table)

    class Warning(OWWidget.Warning):
        all_sets_filtered = Msg('All sets were filtered out.')

    class Error(OWWidget.Error):
        organism_mismatch = Msg('Organism in input data and custom gene sets does not match')
        cant_reach_host = Msg('Host orange.biolab.si is unreachable.')
        cant_load_organisms = Msg('No available organisms, please check your connection.')
        custom_gene_sets_table_format = Msg('Custom gene sets data must have genes represented as rows.')

    def __init__(self):
        super().__init__()
        # OWWidget.__init__(self)
        ConcurrentWidgetMixin.__init__(self)

        # Control area
        box = vBox(self.controlArea, True, margin=0)
        self.gs_selection_component: GeneSetSelection = GeneSetSelection(self, box)
        self.gs_selection_component.selection_changed.connect(self._on_selection_changed)

        # Main area
        self.filter_proxy_model = FilterProxyModel()
        self.filter_proxy_model.setFilterKeyColumn(Header.term)

        self.tree_view = QTreeView()
        self.tree_view.setAlternatingRowColors(True)
        self.tree_view.setSortingEnabled(True)
        self.tree_view.sortByColumn(Header.count, Qt.DescendingOrder)

        self.tree_view.setSelectionMode(QTreeView.ExtendedSelection)
        self.tree_view.setEditTriggers(QTreeView.NoEditTriggers)
        self.tree_view.viewport().setMouseTracking(True)
        self.tree_view.setItemDelegateForColumn(Header.term, LinkStyledItemDelegate(self.tree_view))
        self.tree_view.setItemDelegateForColumn(Header.genes, NumericalColumnDelegate(self))
        self.tree_view.setItemDelegateForColumn(Header.count, NumericalColumnDelegate(self))
        self.tree_view.setModel(self.filter_proxy_model)

        h_layout = QHBoxLayout()
        h_layout.setSpacing(100)
        h_widget = widgetBox(self.mainArea, orientation=h_layout)

        spin(
            h_widget,
            self,
            'min_count',
            0,
            1000,
            label='Count',
            tooltip='Minimum genes count',
            checked='use_min_count',
            callback=self.filter_view,
            callbackOnReturn=True,
            checkCallback=self.filter_view,
        )

        self.line_edit_filter = lineEdit(h_widget, self, 'search_pattern')
        self.line_edit_filter.setPlaceholderText('Filter gene sets ...')
        self.line_edit_filter.textChanged.connect(self.filter_view)

        self.mainArea.layout().addWidget(self.tree_view)
        self.tree_view.header().setSectionResizeMode(QHeaderView.ResizeToContents)

        self.commit_button = auto_commit(self.controlArea, self, 'auto_commit', '&Commit', box=False)

        self.input_data: Optional[Table] = None
        self.num_of_selected_genes: int = 0

    @property
    def tax_id(self) -> Optional[str]:
        if self.input_data:
            return self.input_data.attributes[TableAnnotation.tax_id]

    @property
    def gene_as_attr_name(self) -> Optional[bool]:
        if self.input_data:
            return self.input_data.attributes[TableAnnotation.gene_as_attr_name]

    @property
    def gene_location(self) -> Optional[str]:
        if not self.input_data:
            return

        if self.gene_as_attr_name:
            return self.input_data.attributes[TableAnnotation.gene_id_attribute]
        else:
            return self.input_data.attributes[TableAnnotation.gene_id_column]

    @property
    def input_genes(self) -> Set[str]:
        if not self.input_data:
            return set()

        if self.gene_as_attr_name:
            return {
                str(variable.attributes.get(self.gene_location, '?')) for variable in self.input_data.domain.attributes
            }
        else:
            return {str(g) for g in self.input_data.get_column_view(self.gene_location)[0]}

    def on_partial_result(self, _):
        pass

    def on_done(self, result: Results):
        model = QStandardItemModel()
        for item in result.items:
            model.appendRow(item)

        model.setSortRole(Qt.UserRole)
        model.setHorizontalHeaderLabels(Header.labels())

        self.filter_proxy_model.setSourceModel(model)
        self.tree_view.selectionModel().selectionChanged.connect(self.commit)
        self.filter_view()
        self.update_info_box()

    def on_exception(self, ex):
        # TODO: handle possible exceptions
        raise ex

    def onDeleteWidget(self):
        self.shutdown()
        super().onDeleteWidget()

    def _on_selection_changed(self):
        self.start(run, self.gs_selection_component.gene_sets, self.gs_selection_component.selection, self.input_genes)

    @Inputs.data
    @check_table_annotation
    def set_data(self, input_data: Table):
        self.Outputs.matched_genes.send(None)
        self.input_data = None
        self.num_of_selected_genes = 0

        if input_data:
            self.input_data = input_data
            self.gs_selection_component.initialize(self.tax_id)

        self.update_info_box()

    @Inputs.custom_gene_sets
    def handle_custom_gene_sets_input(self, custom_data):
        self.Outputs.matched_genes.send(None)

        if custom_data:
            self.gs_selection_component.initialize_custom_gene_sets(custom_data)
        else:
            self.gs_selection_component.initialize_custom_gene_sets(None)

        self.update_info_box()

    def commit(self):
        selection_model = self.tree_view.selectionModel()
        self.num_of_selected_genes = 0

        if selection_model:
            selection = selection_model.selectedRows(Header.count)
            self.selected_rows = [self.filter_proxy_model.mapToSource(sel).row() for sel in selection]

            if selection and self.input_genes:
                genes = [model_index.data(Qt.UserRole) for model_index in selection]
                output_genes = list(set.union(*genes))
                self.num_of_selected_genes = len(output_genes)

                if self.gene_as_attr_name:
                    selected = [
                        column
                        for column in self.input_data.domain.attributes
                        if self.gene_location in column.attributes
                        and str(column.attributes[self.gene_location]) in output_genes
                    ]

                    domain = Domain(selected, self.input_data.domain.class_vars, self.input_data.domain.metas)
                    new_data = self.input_data.from_table(domain, self.input_data)
                    self.Outputs.matched_genes.send(new_data)
                else:
                    # create filter from selected column for genes
                    only_known = table_filter.FilterStringList(self.gene_location, output_genes)
                    # apply filter to the data
                    data_table = table_filter.Values([only_known])(self.input_data)
                    self.Outputs.matched_genes.send(data_table)

        self.update_info_box()

    def update_info_box(self):
        input_string = ''
        input_number = ''

        if self.input_genes:
            input_string += '{} unique gene names on input.\n'.format(len(self.input_genes))
            input_number += str(len(self.input_genes))
            self.info.set_output_summary(
                str(self.num_of_selected_genes), '{} genes on output.\n'.format(self.num_of_selected_genes)
            )
        else:
            self.info.set_output_summary(self.info.NoOutput)

        if self.gs_selection_component.data:
            num_of_genes = self.gs_selection_component.num_of_genes
            num_of_sets = self.gs_selection_component.num_of_custom_sets
            input_number += f"{'' if input_number else '0'}|{num_of_genes}"
            input_string += '{} marker genes in {} sets\n'.format(num_of_genes, num_of_sets)

        if not input_number:
            self.info.set_input_summary(self.info.NoInput)
        else:
            self.info.set_input_summary(input_number, input_string)

    def create_filters(self):
        search_term: List[str] = self.search_pattern.lower().strip().split()

        filters = [
            FilterProxyModel.Filter(
                Header.term, Qt.DisplayRole, lambda value: all(fs in value.lower() for fs in search_term)
            )
        ]

        if self.use_min_count:
            filters.append(
                FilterProxyModel.Filter(Header.count, Qt.DisplayRole, lambda value: value >= self.min_count)
            )

        return filters

    def filter_view(self):
        filter_proxy: FilterProxyModel = self.filter_proxy_model
        model: QStandardItemModel = filter_proxy.sourceModel()

        if isinstance(model, QStandardItemModel):

            # apply filtering rules
            filter_proxy.set_filters(self.create_filters())

            if model.rowCount() and not filter_proxy.rowCount():
                self.Warning.all_sets_filtered()
            else:
                self.Warning.clear()

    def sizeHint(self):
        return QSize(800, 600)


if __name__ == "__main__":
    from Orange.widgets.utils.widgetpreview import WidgetPreview

    widget = WidgetPreview(OWGeneSets)
    widget.run()



import io
from enum import IntEnum
from typing import Any, Dict, List, Tuple, Optional, NamedTuple
from datetime import datetime

import numpy as np
import pandas as pd
from resdk.resources.data import Data
from dateutil.relativedelta import relativedelta

from AnyQt.QtCore import (
    Qt,
    QSize,
    QObject,
    QAbstractAnimation,
    QPropertyAnimation,
    QParallelAnimationGroup,
    pyqtSignal,
)
from AnyQt.QtWidgets import (
    QFrame,
    QLabel,
    QDialog,
    QComboBox,
    QLineEdit,
    QTableView,
    QHBoxLayout,
    QHeaderView,
    QPushButton,
    QScrollArea,
    QSizePolicy,
    QToolButton,
    QVBoxLayout,
    QAbstractItemView,
)

from Orange.data import Table, Domain
from Orange.widgets import gui, widget, settings
from Orange.widgets.widget import Msg, Output, StateInfo, OWComponent
from Orange.data.pandas_compat import table_from_frame
from Orange.widgets.credentials import CredentialManager
from Orange.widgets.utils.concurrent import TaskState, ConcurrentWidgetMixin
from Orange.widgets.utils.itemmodels import PyTableModel

from orangecontrib.bioinformatics.resolwe import ResolweAPI, ResolweAuthException, ResolweDataObjectsNotFound, connect
from orangecontrib.bioinformatics.ncbi.gene import GeneMatcher
from orangecontrib.bioinformatics.preprocess import ZScore, LogarithmicScale, QuantileTransform, QuantileNormalization
from orangecontrib.bioinformatics.ncbi.taxonomy import species_name_to_taxid
from orangecontrib.bioinformatics.resolwe.resapi import DEFAULT_URL, RESOLWE_URLS, CREDENTIAL_MANAGER_SERVICE
from orangecontrib.bioinformatics.widgets.utils.data import TableAnnotation
from orangecontrib.bioinformatics.resolwe.sample_metadata import merge_clinical_data

PAST_HOUR = relativedelta(hours=-1)
PAST_DAY = relativedelta(days=-1)
PAST_WEEK = relativedelta(weeks=-1)
PAST_MONTH = relativedelta(months=-1)


class FilterByDateModified(IntEnum):
    any_time = 0
    past_hour = 1
    past_day = 2
    past_week = 3
    past_month = 4

    @staticmethod
    def values():
        now = datetime.now()
        return [None, now + PAST_HOUR, now + PAST_DAY, now + PAST_WEEK, now + PAST_MONTH]

    @staticmethod
    def labels():
        return ['Any time', 'Past hour', 'Past 24 hours', 'Past week', 'Past month']


class SortBy(IntEnum):
    relevance = 0
    name_a_to_z = 1
    name_z_to_a = 2
    contributor_a_to_z = 3
    contributor_z_to_a = 4
    newest_first = 5
    oldest_first = 6

    @staticmethod
    def values():
        return ['', 'name', '-name', 'contributor__last_name', '-contributor__last_name', '-modified', 'modified']

    @staticmethod
    def labels():
        return [
            'Relevance',
            'Name A-Z',
            'Name Z-A',
            'Contributor A-Z',
            'Contributor Z-A',
            'Date modified (newest first)',
            'Date modified (oldest first)',
        ]


class ItemsPerPage(IntEnum):
    min = 0
    med = 1
    max = 2

    @staticmethod
    def values():
        return [20, 50, 100]


class QuantileTransformDist(IntEnum):
    normal = 0
    uniform = 1

    @staticmethod
    def values():
        return ['normal', 'uniform']


class TableHeader(IntEnum):
    id = 0
    slug = 1
    name = 2
    samples = 3
    species = 4
    created = 5
    modified = 6
    contributor = 7
    description = 8
    tags = 9

    @staticmethod
    def labels():
        return [
            'Id',
            'Slug',
            'Name',
            'Samples',
            'Species',
            'Created',
            'Modified',
            'Contributor',
            'Description',
            'Tags',
        ]


class CollapsibleFilterComponent(OWComponent, QObject):
    options_changed = pyqtSignal()

    filter_by_full_text: str = settings.Setting('', schema_only=True)
    filter_by_name: str = settings.Setting('', schema_only=True)
    filter_by_contrib: str = settings.Setting('', schema_only=True)
    filter_by_owner: str = settings.Setting('', schema_only=True)
    filter_by_modified: int = settings.Setting(FilterByDateModified.any_time, schema_only=True)
    sort_by: int = settings.Setting(SortBy.newest_first, schema_only=True)

    FILTER_FULL_TEXT_LABEL = 'Search'
    TOGGLE_BTN_LABEL = 'Narrow your search'
    FILTER_NAME_LABEL = 'Filter by name'
    FILTER_CONTRIB_LABEL = 'Filter by contributor'
    FILTER_OWNER_LABEL = 'Filter by owner'
    FILTER_MODIFIED_LABEL = 'Filter by date modified'
    SORTING_LABEL = 'Sorting'

    def __init__(self, parent_widget, parent_component):
        QObject.__init__(self)
        OWComponent.__init__(self, widget=parent_widget)

        box = gui.widgetBox(parent_component, margin=0)

        self.filter_full_text = gui.lineEdit(
            box,
            self,
            'filter_by_full_text',
            label=self.FILTER_FULL_TEXT_LABEL,
            callback=self.on_filter_full_text_changed,
        )

        self.toggle_animation = QParallelAnimationGroup()
        self.toggle_button = QToolButton()
        self.toggle_button.setCheckable(True)
        self.toggle_button.setChecked(False)
        self.toggle_button.setToolButtonStyle(Qt.ToolButtonTextBesideIcon)
        self.toggle_button.setArrowType(Qt.RightArrow)
        self.toggle_button.setText(self.TOGGLE_BTN_LABEL)
        self.toggle_button.setStyleSheet('QToolButton {border: none; padding-top: 5px; }')
        self.toggle_button.setIconSize(QSize(15, 15))
        self.toggle_button.pressed.connect(self.on_toggle)

        self.collapsible_components = QScrollArea()
        self.collapsible_components.setMaximumHeight(0)
        self.collapsible_components.setMinimumHeight(0)
        self.collapsible_components.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        self.collapsible_components.setFrameShape(QFrame.NoFrame)

        box = gui.widgetBox(parent_component, margin=0)
        box.layout().addWidget(self.toggle_button)
        box.layout().addWidget(self.collapsible_components)

        self.toggle_animation.addAnimation(QPropertyAnimation(box, b"minimumHeight"))
        self.toggle_animation.addAnimation(QPropertyAnimation(box, b"maximumHeight"))
        self.toggle_animation.addAnimation(QPropertyAnimation(self.collapsible_components, b"maximumHeight"))

        layout = QHBoxLayout()
        left_box = gui.widgetBox(None, self, margin=0, flat=True)
        mid_box = gui.widgetBox(None, self, margin=0, flat=True)
        right_box = gui.widgetBox(None, self, margin=0, flat=True)

        self.filter_name = gui.lineEdit(
            left_box, self, 'filter_by_name', label=self.FILTER_NAME_LABEL, callback=self.on_filter_changed, addSpace=5
        )
        self.filter_contrib = gui.lineEdit(
            mid_box,
            self,
            'filter_by_contrib',
            label=self.FILTER_CONTRIB_LABEL,
            callback=self.on_filter_changed,
            addSpace=5,
        )
        self.filter_owner = gui.lineEdit(
            right_box,
            self,
            'filter_by_owner',
            label=self.FILTER_OWNER_LABEL,
            callback=self.on_filter_changed,
            addSpace=5,
        )
        self.filter_modified = gui.comboBox(
            left_box,
            self,
            'filter_by_modified',
            label=self.FILTER_MODIFIED_LABEL,
            callback=self.on_filter_changed,
            items=FilterByDateModified.labels(),
        )
        self.sorting = gui.comboBox(
            mid_box, self, 'sort_by', label=self.SORTING_LABEL, callback=self.on_filter_changed, items=SortBy.labels()
        )

        gui.rubber(left_box)
        gui.rubber(mid_box)
        gui.rubber(right_box)
        layout.addWidget(left_box)
        layout.addWidget(mid_box)
        layout.addWidget(right_box)
        self.collapsible_components.setLayout(layout)

        collapsed_height = box.layout().sizeHint().height() - self.collapsible_components.maximumHeight()
        content_height = layout.sizeHint().height()

        for i in range(self.toggle_animation.animationCount()):
            animation = self.toggle_animation.animationAt(i)
            animation.setDuration(100)
            animation.setStartValue(collapsed_height)
            animation.setEndValue(collapsed_height + content_height)

        content_animation = self.toggle_animation.animationAt(self.toggle_animation.animationCount() - 1)
        content_animation.setDuration(100)
        content_animation.setStartValue(0)
        content_animation.setEndValue(content_height)

    def on_toggle(self):
        """ Start animation """
        checked = self.toggle_button.isChecked()
        self.toggle_button.setArrowType(Qt.DownArrow if not checked else Qt.RightArrow)
        self.toggle_animation.setDirection(QAbstractAnimation.Forward if not checked else QAbstractAnimation.Backward)
        self.toggle_animation.start()

    def on_filter_full_text_changed(self):
        self.sort_by = SortBy.relevance if self.filter_full_text else SortBy.newest_first
        self.on_filter_changed()

    def on_filter_changed(self):
        self.options_changed.emit()


class PaginationComponent(OWComponent, QObject):
    options_changed = pyqtSignal()

    items_per_page: int = settings.Setting(ItemsPerPage.min, schema_only=True)
    current_page: int = settings.Setting(1, schema_only=True)
    offset: int = settings.Setting(0, schema_only=True)

    PAGE_LIMIT_LABEL = 'Items per page  '

    def __init__(self, parent_widget, parent_component):
        QObject.__init__(self)
        OWComponent.__init__(self, widget=parent_widget)

        box = gui.widgetBox(parent_component, orientation=Qt.Horizontal)

        self.page_limit = gui.radioButtons(
            box,
            self,
            'items_per_page',
            [str(val) for val in ItemsPerPage.values()],
            orientation=Qt.Horizontal,
            callback=self.on_limit_changed,
            label=self.PAGE_LIMIT_LABEL,
        )

        self.page_left_btn = QToolButton()
        self.page_left_btn.setStyleSheet('QToolButton {border: none;}')
        self.page_left_btn.setArrowType(Qt.LeftArrow)
        self.page_left_btn.pressed.connect(self.left_btn_pressed)

        self.offset_label = gui.label(None, self, str(self.current_page), labelWidth=15)
        self.offset_label.setAlignment(Qt.AlignCenter)

        self.page_right_btn = QToolButton()
        self.page_right_btn.setStyleSheet('QToolButton {border: none;}')
        self.page_right_btn.setArrowType(Qt.RightArrow)
        self.page_right_btn.pressed.connect(self.right_btn_pressed)

        parent_widget.pagination_availability.connect(self._handle_paginate_buttons)

        box.layout().addStretch(1)
        box.layout().addWidget(self.page_left_btn)
        box.layout().addWidget(self.offset_label)
        box.layout().addWidget(self.page_right_btn)

    def _handle_paginate_buttons(self, next_page: bool, previous_page: bool):
        self.page_left_btn.setEnabled(previous_page)
        self.page_right_btn.setEnabled(next_page)

    def reset_pagination(self):
        self.offset = 0
        self.current_page = 1
        self.offset_label.setText(str(self.current_page))

    def left_btn_pressed(self):
        new_offset = self.offset - ItemsPerPage.values()[self.items_per_page]

        if new_offset >= 0:
            self.current_page -= 1
            self.offset = new_offset
            self.offset_label.setText(str(self.current_page))

        self.options_changed.emit()

    def right_btn_pressed(self):
        self.current_page += 1
        self.offset += ItemsPerPage.values()[self.items_per_page]
        self.offset_label.setText(str(self.current_page))
        self.options_changed.emit()

    def on_limit_changed(self):
        self.reset_pagination()
        self.options_changed.emit()


class NormalizationComponent(OWComponent, QObject):
    """ Gene expression normalization component """

    options_changed = pyqtSignal()

    quantile_norm: bool
    quantile_norm = settings.Setting(False, schema_only=True)

    log_norm: bool
    log_norm = settings.Setting(True, schema_only=True)

    z_score_norm: bool
    z_score_norm = settings.Setting(False, schema_only=True)

    z_score_axis: int
    z_score_axis = settings.Setting(0, schema_only=True)

    quantile_transform: bool
    quantile_transform = settings.Setting(False, schema_only=True)

    quantile_transform_axis: int
    quantile_transform_axis = settings.Setting(0, schema_only=True)

    quantile_transform_dist: int
    quantile_transform_dist = settings.Setting(0, schema_only=True)

    BOX_TITLE = 'Normalization'
    QUANTILE_NORM_LABEL = 'Quantile normalization'
    LOG_NORM_LABEL = 'Log2(x+1)'
    Z_SCORE_LABEL = 'Z-score'
    QUANTILE_TRANSFORM_LABEL = 'Quantile transform'

    def __init__(self, parent_widget, parent_component):
        QObject.__init__(self)
        OWComponent.__init__(self, widget=parent_widget)

        box = gui.widgetBox(parent_component, self.BOX_TITLE, margin=3)
        gui.checkBox(box, self, 'quantile_norm', self.QUANTILE_NORM_LABEL, callback=self.options_changed.emit)
        gui.checkBox(box, self, 'log_norm', self.LOG_NORM_LABEL, callback=self.options_changed.emit)
        gui.checkBox(box, self, 'z_score_norm', self.Z_SCORE_LABEL, callback=self.on_z_score_selected)
        self.z_score_axis_btn = gui.radioButtons(
            gui.indentedBox(box),
            self,
            'z_score_axis',
            btnLabels=['columns', 'rows'],
            callback=self.options_changed.emit,
            orientation=Qt.Horizontal,
        )
        self.z_score_axis_btn.setHidden(not bool(self.z_score_norm))

        gui.checkBox(
            box,
            self,
            'quantile_transform',
            self.QUANTILE_TRANSFORM_LABEL,
            callback=self.on_quantile_transform_selected,
        )
        self.quantile_transform_axis_btn = gui.radioButtons(
            gui.indentedBox(box),
            self,
            'quantile_transform_axis',
            btnLabels=['columns', 'rows'],
            callback=self.options_changed.emit,
            orientation=Qt.Horizontal,
        )
        self.quantile_transform_dist_btn = gui.radioButtons(
            gui.indentedBox(box),
            self,
            'quantile_transform_dist',
            btnLabels=QuantileTransformDist.values(),
            callback=self.options_changed.emit,
            orientation=Qt.Horizontal,
        )

        self.quantile_transform_axis_btn.setHidden(not bool(self.quantile_transform))
        self.quantile_transform_dist_btn.setHidden(not bool(self.quantile_transform))

    def on_z_score_selected(self):
        self.z_score_axis_btn.setHidden(not bool(self.z_score_norm))
        self.options_changed.emit()

    def on_quantile_transform_selected(self):
        self.quantile_transform_axis_btn.setHidden(not bool(self.quantile_transform))
        self.quantile_transform_dist_btn.setHidden(not bool(self.quantile_transform))
        self.options_changed.emit()


class SignInForm(QDialog):
    def __init__(self, flags, *args, **kwargs):
        super().__init__(flags, *args, **kwargs)
        self.cm: CredentialManager = CredentialManager(CREDENTIAL_MANAGER_SERVICE)

        self.setWindowTitle('Sign in')
        self.setFixedSize(400, 250)

        self.server_cb_label = QLabel('Server *')
        self.server_cb = QComboBox(self)
        self.server_cb.addItems(RESOLWE_URLS)
        self.server_cb.setEditable(True)

        self.username_label = QLabel('Username *')
        self.username_line_edit = QLineEdit(self)
        self.username_line_edit.setPlaceholderText('Enter correct username')
        self.username_line_edit.returnPressed.connect(self.sign_in)
        self.username_line_edit.textChanged.connect(self.handle_sign_in_btn)

        self.password_label = QLabel('Password *')
        self.password_line_edit = QLineEdit(self)
        self.password_line_edit.setPlaceholderText('Enter correct password')
        self.password_line_edit.returnPressed.connect(self.sign_in)
        self.password_line_edit.textChanged.connect(self.handle_sign_in_btn)
        self.password_line_edit.setEchoMode(QLineEdit.Password)

        self.sign_in_btn = QPushButton('Sign in', self)
        self.sign_in_btn.setDisabled(True)
        self.sign_in_btn.clicked.connect(self.sign_in)

        self.error_msg = QLabel('Unable to log in with provided credentials.')
        self.error_msg.setStyleSheet('color:red')
        self.error_msg.hide()

        layout = QVBoxLayout(self)
        layout.addWidget(self.server_cb_label)
        layout.addWidget(self.server_cb)
        layout.addWidget(self.username_label)
        layout.addWidget(self.username_line_edit)
        layout.addWidget(self.password_label)
        layout.addWidget(self.password_line_edit)
        layout.addWidget(self.error_msg)
        layout.addStretch()
        layout.addWidget(self.sign_in_btn)

        self.resolwe_instance = None

    def handle_sign_in_btn(self):
        self.sign_in_btn.setEnabled(
            True if self.username_line_edit.text() and self.password_line_edit.text() else False
        )

    def sign_in(self):
        self.server_cb_label.setStyleSheet(None)
        self.username_label.setStyleSheet(None)
        self.password_label.setStyleSheet(None)
        self.error_msg.hide()

        server = self.server_cb.currentText()
        username = self.cm.username if self.cm.username else self.username_line_edit.text()
        password = self.cm.password if self.cm.password else self.password_line_edit.text()

        if not server:
            self.server_cb_label.setStyleSheet('color:red')
            return

        if not username:
            self.username_label.setStyleSheet('color:red')
            return

        if not password:
            self.password_label.setStyleSheet('color:red')
            return

        try:
            self.resolwe_instance = connect(username, password, url=server)
        except ResolweAuthException:
            self.error_msg.show()
            return

        self.cm.username = username
        self.cm.password = password
        self.accept()


class GenialisExpressionsModel(PyTableModel):
    def __init__(self, parent_widget, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.parent = parent_widget

    def flags(self, index):
        """
        Disable the row selection by clicking on the first column.
        """
        return Qt.ItemIsEnabled if index.column() == 0 else Qt.ItemIsEnabled | Qt.ItemIsSelectable

    def data(self, index, role=Qt.DisplayRole):
        if not index.isValid():
            return

        row, column = self.mapToSourceRows(index.row()), index.column()

        role_value = self._roleData.get(row, {}).get(column, {}).get(role)
        if role_value is not None:
            return role_value

        try:
            value = self[row][column]
            tag = self[row][TableHeader.tags]
            slug = self[row][TableHeader.slug]
            url = f'{self.parent.res.url}/{tag}/search/collection/{slug}'
        except IndexError:
            return

        if role == Qt.DisplayRole:
            if isinstance(value, datetime):
                return str(value.date().strftime('%m/%d/%Y'))
            elif column == TableHeader.description:
                if value:
                    return f'{str(value[:100])} ...'
            return str(value)

        elif role == Qt.ToolTipRole:
            if isinstance(value, datetime):
                return str(value.strftime('%b %d, %Y %H:%M'))
            if tag and column == TableHeader.id:
                return url
            return str(value)

        # elif role == Qt.TextAlignmentRole and isinstance(value, Number):
        #     return Qt.AlignRight | Qt.AlignVCenter

        elif role == gui.LinkRole:
            if tag:
                return url

    def set_data(self, collections: List[Dict[str, str]], col_to_species: Dict[str, str]):
        def model_row(collection: Dict[str, str]) -> List[Any]:
            first_name = collection.get('contributor', {}).get('first_name', '')
            last_name = collection.get('contributor', {}).get('last_name', '')
            user_name = collection.get('contributor', {}).get('username', '')
            contributor = f'{first_name} {last_name}'.strip()

            tags = collection.get('tags', [])
            tags = [tag for tag in tags if 'community' in tag]
            if len(tags) == 1:
                _, tag = tags[0].split(':')
            else:
                tag = ''

            return [
                collection['id'],
                collection.get('slug', ''),
                collection.get('name', ''),
                collection.get('entity_count', ''),
                col_to_species.get(collection['id'], ''),
                datetime.strptime(collection.get('created', ''), '%Y-%m-%dT%H:%M:%S.%f%z'),
                datetime.strptime(collection.get('modified', ''), '%Y-%m-%dT%H:%M:%S.%f%z'),
                contributor if contributor else user_name,
                collection.get('description', ''),
                tag,
            ]

        self.wrap([model_row(result) for result in collections])


class Expression(NamedTuple):
    type: str
    name: str


class Process(NamedTuple):
    type: str
    name: str


class InputAnnotation(NamedTuple):
    source: str
    species: str
    build: str


class DataOutputOptions(NamedTuple):
    expression: Tuple[Expression]
    process: Tuple[Process]
    input_annotation: Tuple[InputAnnotation]


def available_data_output_options(data_objects: List[Data]) -> DataOutputOptions:
    """
    Traverse the data objects in the selected collection and store the
    information regarding available expression types, process types and
    input annotations.

    The method returns a named tuple (`DataOutputOptions`) which used for
    creating radio buttons in the control area.
    """

    expression_types = sorted({data.output['exp_type'] for data in data_objects})
    expression_types = (Expression('rc', 'Read Counts'),) + tuple(
        Expression(exp_type, exp_type) for exp_type in expression_types
    )

    process_types = sorted({(data.process.type, data.process.name) for data in data_objects})
    process_types = tuple(Process(proc_type, proc_name) for proc_type, proc_name in process_types)

    input_annotations = sorted(
        {(data.output['source'], data.output['species'], data.output['build']) for data in data_objects}
    )
    input_annotations = tuple(InputAnnotation(source, species, build) for source, species, build in input_annotations)

    return DataOutputOptions(expression=expression_types, process=process_types, input_annotation=input_annotations)


class OWGenialisExpressions(widget.OWWidget, ConcurrentWidgetMixin):
    name = 'Genialis Expressions'
    priority = 30
    want_main_area = True
    want_control_area = True
    icon = '../widgets/icons/OWGenialisExpressions.svg'

    pagination_availability = pyqtSignal(bool, bool)

    norm_component = settings.SettingProvider(NormalizationComponent)
    pagination_component = settings.SettingProvider(PaginationComponent)
    filter_component = settings.SettingProvider(CollapsibleFilterComponent)

    exp_type: int
    exp_type = settings.Setting(0, schema_only=True)

    proc_type: int
    proc_type = settings.Setting(0, schema_only=True)

    input_annotation: int
    input_annotation = settings.Setting(0, schema_only=True)

    selected_orange_data_object: int
    selected_orange_data_object = settings.Setting(0, schema_only=True)

    auto_commit: bool
    auto_commit = settings.Setting(False, schema_only=True)

    class Outputs:
        table = Output('Expressions', Table)

    class Warning(widget.OWWidget.Warning):
        no_expressions = Msg('Expression data objects not found.')
        no_data_objects = Msg('No expression data matches the selected options.')
        unexpected_feature_type = Msg('Can not import expression data, unexpected feature type "{}".')
        multiple_feature_type = Msg('Can not import expression data, multiple feature types found.')

    def __init__(self):
        super().__init__()
        ConcurrentWidgetMixin.__init__(self)

        self._res: Optional[ResolweAPI] = None

        # Store collection ID from currently selected row
        self.selected_collection_id: Optional[str] = None
        # Store data output options
        self.data_output_options: Optional[DataOutputOptions] = None
        # List of all available data objects (expressions)
        self.data_objects: Optional[List[Data]] = None
        # Subset of data objects based on DataOutputOptions
        self.filtered_data_objects: Optional[List[Data]] = None
        # List of orange data objects (clinical data)
        self.orange_data_objects: Dict[int, Data] = {}
        # Cache output data table
        self.data_table: Optional[Table] = None
        # Cache clinical metadata
        self.clinical_metadata: Optional[Table] = None

        # Control area
        self.info_box = gui.widgetLabel(gui.widgetBox(self.controlArea, "Info", margin=3), 'No data on output.')

        self.exp_type_combo = gui.comboBox(
            self.controlArea, self, 'exp_type', label='Expression Type', callback=self.on_output_option_changed
        )
        self.proc_type_combo = gui.comboBox(
            self.controlArea, self, 'proc_type', label='Process Name', callback=self.on_output_option_changed
        )
        self.input_anno_combo = gui.comboBox(
            self.controlArea,
            self,
            'input_annotation',
            label='Expression source',
            callback=self.on_output_option_changed,
        )
        self.orange_data_objects_combo = gui.comboBox(
            self.controlArea,
            self,
            'selected_orange_data_object',
            label='Clinical Data',
            callback=self.on_clinical_data_changed,
        )

        self.norm_component = NormalizationComponent(self, self.controlArea)
        self.norm_component.options_changed.connect(self.on_normalization_changed)

        gui.rubber(self.controlArea)
        box = gui.widgetBox(self.controlArea, 'Sign in')
        self.user_info = gui.label(box, self, '')
        self.server_info = gui.label(box, self, '')

        box = gui.widgetBox(box, orientation=Qt.Horizontal)
        self.sign_in_btn = gui.button(box, self, 'Sign In', callback=self.sign_in, autoDefault=False)
        self.sign_out_btn = gui.button(box, self, 'Sign Out', callback=self.sign_out, autoDefault=False)

        self.commit_button = gui.auto_commit(self.controlArea, self, 'auto_commit', '&Commit', box=False)
        self.commit_button.button.setAutoDefault(False)

        # Main area
        self.table_view = QTableView()
        self.table_view.setAlternatingRowColors(True)
        self.table_view.viewport().setMouseTracking(True)
        self.table_view.setShowGrid(False)
        self.table_view.verticalHeader().hide()
        self.table_view.horizontalHeader().setSectionResizeMode(QHeaderView.ResizeToContents)
        self.table_view.horizontalHeader().setStretchLastSection(True)
        self.table_view.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.table_view.setSelectionMode(QAbstractItemView.SingleSelection)
        # self.table_view.setStyleSheet('QTableView::item:selected{background-color: palette(highlight); color: palette(highlightedText);};')

        self.model = GenialisExpressionsModel(self)
        self.model.setHorizontalHeaderLabels(TableHeader.labels())
        self.table_view.setModel(self.model)
        self.table_view.selectionModel().selectionChanged.connect(self.on_selection_changed)

        self.filter_component = CollapsibleFilterComponent(self, self.mainArea)
        self.filter_component.options_changed.connect(self.on_filter_changed)
        self.mainArea.layout().addWidget(self.table_view)
        self.pagination_component = PaginationComponent(self, self.mainArea)
        self.pagination_component.options_changed.connect(self.update_collections_view)

        self.sign_in(silent=True)

    @property
    def res(self):
        return self._res

    @res.setter
    def res(self, value: ResolweAPI):
        if isinstance(value, ResolweAPI):
            self._res = value
            self.update_user_status()
            self.update_collections_view()
            self.__invalidate()
            self.Outputs.table.send(None)

    def __invalidate(self):
        self.data_objects = None
        self.data_table = None
        self.selected_collection_id = None
        self.clinical_metadata = None
        self.orange_data_objects = {}

        self.data_output_options = None
        self.exp_type_combo.clear()
        self.proc_type_combo.clear()
        self.input_anno_combo.clear()
        self.orange_data_objects_combo.clear()

        self.Outputs.table.send(None)
        self.Warning.no_expressions.clear()
        self.Warning.multiple_feature_type.clear()
        self.Warning.unexpected_feature_type.clear()
        self.Warning.no_data_objects.clear()
        self.info.set_output_summary(StateInfo.NoOutput)
        self.update_info_box()

    def update_user_status(self):
        user = self.res.get_currently_logged_user()

        if user:
            user_info = f"{user[0].get('first_name', '')} {user[0].get('last_name', '')}".strip()
            user_info = f"User: {user_info if user_info else user[0].get('username', '')}"
            self.sign_in_btn.setEnabled(False)
            self.sign_out_btn.setEnabled(True)
        else:
            user_info = 'User: Anonymous'
            self.sign_in_btn.setEnabled(True)
            self.sign_out_btn.setEnabled(False)

        self.user_info.setText(user_info)
        self.server_info.setText(f'Server: {self.res.url[8:]}')

    def update_info_box(self):

        if self.data_table:
            total_genes = len(self.data_table.domain.attributes)
            known_genes = len([col for col in self.data_table.domain.attributes if len(col.attributes)])

            info_text = (
                '{} genes on output\n'
                '{} genes match Entrez database\n'
                '{} genes with match conflicts\n'.format(total_genes, known_genes, total_genes - known_genes)
            )

        else:
            info_text = 'No data on output.'

        self.info_box.setText(info_text)

    def sign_in(self, silent=False):
        dialog = SignInForm(self)

        if silent:
            dialog.sign_in()
            if dialog.resolwe_instance is not None:
                self.res = dialog.resolwe_instance
            else:
                self.res = connect(url=DEFAULT_URL)

        if not silent and dialog.exec_():
            self.res = dialog.resolwe_instance

    def sign_out(self):
        # Use public credentials when user signs out
        self.res = connect(url=DEFAULT_URL)
        # Remove username and
        cm = CredentialManager(CREDENTIAL_MANAGER_SERVICE)
        del cm.username
        del cm.password

    def on_filter_changed(self):
        self.pagination_component.reset_pagination()
        self.update_collections_view()

    def get_query_parameters(self) -> Dict[str, str]:
        params = {
            'limit': ItemsPerPage.values()[self.pagination_component.items_per_page],
            'offset': self.pagination_component.offset,
            'ordering': SortBy.values()[self.filter_component.sort_by],
        }

        if self.filter_component.filter_by_full_text:
            params.update({'text': self.filter_component.filter_by_full_text})

        if self.filter_component.filter_by_name:
            params.update({'name__icontains': self.filter_component.filter_by_name})

        if self.filter_component.filter_by_contrib:
            params.update({'contributor_name': self.filter_component.filter_by_contrib})

        if self.filter_component.filter_by_owner:
            params.update({'owners_name': self.filter_component.filter_by_owner})

        last_modified = FilterByDateModified.values()[self.filter_component.filter_by_modified]
        if last_modified:
            params.update({'modified__gte': last_modified.isoformat()})

        return params

    def get_collections(self) -> Tuple[Dict[str, str], Dict[str, str]]:
        # Get response from the server
        collections = self.res.get_collections(**self.get_query_parameters())
        # Loop trough collections and store ids
        collection_ids = [collection['id'] for collection in collections.get('results', [])]
        # Get species by collection ids
        collection_to_species = self.res.get_species(collection_ids)

        return collections, collection_to_species

    def update_collections_view(self):
        collections, collection_to_species = self.get_collections()

        # Pass the results to data model
        self.model.set_data(collections.get('results', []), collection_to_species)
        self.table_view.setItemDelegateForColumn(TableHeader.id, gui.LinkStyledItemDelegate(self.table_view))
        self.table_view.setColumnHidden(TableHeader.slug, True)
        self.table_view.setColumnHidden(TableHeader.tags, True)

        # Check pagination parameters and emit pagination_availability signal
        next_page = True if collections.get('next') else False
        previous_page = True if collections.get('previous') else False
        self.pagination_availability.emit(next_page, previous_page)

    def normalize(self, table: Table) -> Optional[Table]:
        if not table:
            return

        if self.norm_component.quantile_norm:
            table = QuantileNormalization()(table)

        if self.norm_component.log_norm:
            table = LogarithmicScale()(table)

        if self.norm_component.z_score_norm:
            table = ZScore(axis=self.norm_component.z_score_axis)(table)

        if self.norm_component.quantile_transform:
            axis = self.norm_component.quantile_transform_axis
            quantiles = table.X.shape[int(not axis)]
            distribution = QuantileTransformDist.values()[self.norm_component.quantile_transform_dist]
            table = QuantileTransform(axis=axis, n_quantiles=quantiles, output_distribution=distribution)(table)

        return table

    def commit(self):
        self.Warning.no_data_objects.clear()
        self.cancel()

        if not self.filtered_data_objects:
            self.Warning.no_data_objects()
            self.Outputs.table.send(None)
            self.info.set_output_summary(StateInfo.NoOutput)
            return

        self.start(self.runner)

    def filter_data_objects(self):
        options = self.data_output_options
        exp_type = options.expression[self.exp_type].type
        proc_type = options.process[self.proc_type].type
        source = options.input_annotation[self.input_annotation].source
        species = options.input_annotation[self.input_annotation].species
        build = options.input_annotation[self.input_annotation].build

        def is_selected(obj):
            return all(
                [
                    obj.process.type == proc_type,
                    obj.output['source'] == source,
                    obj.output['species'] == species,
                    obj.output['build'] == build,
                    obj.output['exp_type'] == exp_type if exp_type != 'rc' else True,
                ]
            )

        return [data_object for data_object in self.data_objects if is_selected(data_object)]

    def fetch_clinical_metadata(self):
        data_object = self.orange_data_objects.get(self.selected_orange_data_object, None)
        if data_object is not None:
            return self.res.get_clinical_metadata(data_object.id, data_object.output['table']['file'])

    def on_output_option_changed(self):
        self.data_table = None
        self.filtered_data_objects = self.filter_data_objects()
        self.commit()

    def on_clinical_data_changed(self):
        self.clinical_metadata = self.fetch_clinical_metadata()
        self.commit()

    def on_normalization_changed(self):
        self.commit()

    def on_selection_changed(self):
        self.__invalidate()

        collection_id: str = self.get_selected_row_data(TableHeader.id)
        if not collection_id:
            return

        self.selected_collection_id = collection_id
        self.data_objects = self.res.get_expression_data_objects(collection_id)
        self.orange_data_objects = {
            idx: data for idx, data in enumerate(self.res.get_orange_data_objects(collection_id))
        }

        self.data_output_options = available_data_output_options(self.data_objects)

        self.exp_type_combo.addItems(exp_name for _, exp_name in self.data_output_options.expression)
        if self.exp_type >= len(self.data_output_options.expression):
            self.exp_type = 0
        self.exp_type_combo.setCurrentIndex(self.exp_type)

        self.proc_type_combo.addItems(proc_name for _, proc_name in self.data_output_options.process)
        if self.proc_type >= len(self.data_output_options.process):
            self.proc_type = 0
        self.proc_type_combo.setCurrentIndex(self.proc_type)

        self.input_anno_combo.addItems(
            f'{species}, {build}' for _, species, build in self.data_output_options.input_annotation
        )
        if self.input_annotation >= len(self.data_output_options.input_annotation):
            self.input_annotation = 0
        self.input_anno_combo.setCurrentIndex(self.input_annotation)

        self.orange_data_objects_combo.addItems(data.name for data in self.orange_data_objects.values())
        self.clinical_metadata = self.fetch_clinical_metadata()

        if not self.data_objects:
            self.Warning.no_expressions()
            return

        # Note: This here is to handle an edge case where we get
        #       different 'feature_type' data object in a collection.
        #       For now we raise a warning, but in the future we should
        #       discuss about how to properly handle different types of features.
        feature_types = {data.output['feature_type'] for data in self.data_objects}

        if len(feature_types) == 1 and 'gene' not in feature_types:
            self.Warning.unexpected_feature_type(feature_types.pop())
            self.data_objects = []
            return

        if len(feature_types) > 1:
            self.Warning.multiple_feature_type()
            self.data_objects = []
            return

        self.on_output_option_changed()

    def get_selected_row_data(self, column: int) -> Optional[str]:
        selection_model = self.table_view.selectionModel()
        rows = selection_model.selectedRows(column=column)
        if not rows:
            return

        return rows[0].data()

    def on_done(self, table: Table):
        if table:
            samples, genes = table.X.shape
            self.info.set_output_summary(f'Samples: {samples} Genes: {genes}')
            self.update_info_box()
            self.Outputs.table.send(table)

    def on_exception(self, ex):
        if isinstance(ex, ResolweDataObjectsNotFound):
            self.Warning.no_data_objects()
            self.Outputs.table.send(None)
            self.data_table = None
            self.info.set_output_summary(StateInfo.NoOutput)
            self.update_info_box()
        else:
            raise ex

    def on_partial_result(self, result: Any) -> None:
        pass

    def onDeleteWidget(self):
        self.shutdown()
        super().onDeleteWidget()

    def sizeHint(self):
        return QSize(1280, 620)

    def runner(self, state: TaskState) -> Table:
        data_objects = self.filtered_data_objects
        exp_type = self.data_output_options.expression[self.exp_type].type
        species = self.data_output_options.input_annotation[self.input_annotation].species
        output_field = 'exp' if exp_type != 'rc' else exp_type
        collection_id = self.selected_collection_id

        table = self.data_table
        step, steps = 0, 4

        def set_progress():
            nonlocal step
            step += 1
            state.set_progress_value(100 * (step / steps))

        if not table:
            steps += len(data_objects)
            state.set_status('Downloading ...')
            data_frames = []

            for data_object in data_objects:
                set_progress()

                response = self.res.download(data_object.id, data_object.output[output_field]['file'])
                with io.BytesIO() as f:
                    f.write(response.content)
                    f.seek(0)
                    # expressions to data frame
                    df = pd.read_csv(f, sep='\t', compression='gzip')
                    df = df.set_index('Gene').T.reset_index(drop=True)
                    data_frames.append(df)

            state.set_status('Concatenating samples ...')
            df = pd.concat(data_frames, axis=0)

            state.set_status('To data table ...')
            table = table_from_frame(df)
            set_progress()

            state.set_status('Matching genes ...')
            tax_id = species_name_to_taxid(species)
            gm = GeneMatcher(tax_id)
            table = gm.match_table_attributes(table, rename=True)
            table.attributes[TableAnnotation.tax_id] = tax_id
            table.attributes[TableAnnotation.gene_as_attr_name] = True
            table.attributes[TableAnnotation.gene_id_attribute] = 'Entrez ID'

            self.data_table = table
            set_progress()

        state.set_status('Adding metadata ...')
        metadata = self.res.aggregate_sample_metadata([data.sample for data in data_objects], collection_id)
        domain = Domain(table.domain.attributes, table.domain.class_vars, list(metadata.keys()))
        table = table.transform(domain)
        for key, value in metadata.items():
            table[:, key] = np.reshape(value, (-1, 1))

        if self.clinical_metadata:
            table = merge_clinical_data(table, self.clinical_metadata)
        set_progress()

        state.set_status('Normalizing ...')
        table = self.normalize(table)
        set_progress()

        return table


if __name__ == "__main__":
    from orangewidget.utils.widgetpreview import WidgetPreview

    WidgetPreview(OWGenialisExpressions).run()



""" Utils """
import io
import os
import gzip
import json

from Orange.data import Table, Domain, TimeVariable, StringVariable, ContinuousVariable

from orangecontrib.bioinformatics.utils import local_cache

#  Support cache with requests_cache module
LOCAL_CACHE_DIR = os.path.join(local_cache, 'resolwe')
try:
    os.makedirs(LOCAL_CACHE_DIR)
except OSError:
    pass

CACHE_BACKEND = 'sqlite'
GENAPI_CACHE = os.path.join(LOCAL_CACHE_DIR, 'GenAPI_cache')
RESOLWEAPI_CACHE = os.path.join(LOCAL_CACHE_DIR, 'ResolweAPI_cache')


def transpose_table(table):
    """Transpose the rows and columns of the table.

    Args:
        table: Data in :obj:`Orange.data.Table`

    Returns:
         Transposed :obj:`Orange.data.Table`. (Genes as columns)
    """

    # TODO: remove this and use Orange.data.Table.transpose

    attrs = table.domain.attributes
    attr = [ContinuousVariable.make(ex['Gene'].value) for ex in table]
    #  Set metas
    new_metas = [
        StringVariable.make(name) if name != 'Time' else TimeVariable.make(name)
        for name in sorted(table.domain.variables[0].attributes.keys())
    ]
    domain = Domain(attr, metas=new_metas)
    meta_values = [[exp.attributes[var.name] for var in domain.metas] for exp in attrs]

    return Table(domain, table.X.transpose(), metas=meta_values)


def etc_to_table(etc_json, time_var=False):
    """Converts data from Json to :obj:`Orange.data.table`

    Args:
        etc_json (dict): Data in json like format from genesis
        time_var (bool): Create column of time points. Default is set to False.

    Returns:
        :obj:`Orange.data.Table`
    """

    variables = []
    time_point = 1
    for time in etc_json['etc']['timePoints']:
        var = ContinuousVariable('TP ' + str(time_point))
        var.attributes['Time'] = str(time)
        variables.append(var)
        time_point += 1

    meta_attr = StringVariable.make('Gene')
    domain = Domain(variables, metas=[meta_attr])

    table = []
    for row in etc_json['etc']['genes']:
        gene_expression = list(etc_json['etc']['genes'][row])
        gene_expression.append(row)
        table.append(gene_expression)

    orange_table = Table.from_list(domain, table)

    if time_var:
        orange_table = transpose_table(orange_table)

    return orange_table


def response_to_json(response):
    if not response.ok:
        response.raise_for_status()

    response_gzipped = io.BytesIO(response.content)
    response_content = io.TextIOWrapper(gzip.GzipFile(fileobj=response_gzipped), encoding='utf-8')

    try:
        json_object = json.load(response_content)
    except ValueError as e:
        raise ValueError('Downloaded data is not a valid JSON') from e

    return json_object



import json
from typing import Dict, List, Tuple, Optional
from collections import Counter, defaultdict

import numpy as np
from resdk.resources.data import Sample
from resdk.resources.relation import Relation

from Orange.data import Table, Domain, Variable, TimeVariable, DiscreteVariable
from Orange.data.io_util import guess_data_type


def flatten_descriptor(descriptor: Dict[str, dict]) -> Dict[str, dict]:
    """
    Flattens sample descriptor one level deep. Adjust this
    method if there is a need to support descriptors with
    multiple nested choices.

    For example:
        {
          'general': { ... },
          'experiment': {'chip_seq': { ... }, ...}
        }
    Is split into:
        {
          'general.field_name1': { ... },
          'experiment.field_name2': { ... },
          'chip_seq.field_name3': { ... },
          ...
        }

    """
    flat_descriptor = {}
    for section, values in descriptor.items():
        for name, field_value in values.items():
            if isinstance(field_value, dict):
                # its a subsection
                flat_descriptor.update({f'{name}.{key}': val for key, val in field_value.items()})
            else:
                flat_descriptor.update({f'{section}.{name}': field_value})
    return flat_descriptor


def flatten_descriptor_schema(schema: List[dict]) -> Tuple[Dict[str, dict], Dict[str, dict]]:
    """
    Flattens descriptor schema one level deep. Adjust this
    method if there is a need to support schemas which have
    multiple layers of nested subgroups.

    For example:
        [
          {'name': 'general', 'group': [ ... ]},
          {'name': 'experiment', 'group: [ {'name': 'chip_seq', 'group': [ ... ]}, ... ]}
        ]
    Is split into:
        {
          'general.field_name1': { ... },
          'experiment.field_name2': { ... },
          'chip_seq.field_name3': { ... },
          ...
        }

    """

    flat_schema = {}
    section_name_to_label = {}
    for section in schema:
        section_name_to_label.update({section['name']: section['label']})

        for field in section['group']:
            if 'group' in field:
                # field is a subgroup
                flat_schema.update({f"{field['name']}.{sub_field['name']}": sub_field for sub_field in field['group']})
                section_name_to_label.update({sub_field['name']: sub_field['label'] for sub_field in field['group']})
            else:
                flat_schema.update({f"{section['name']}.{field['name']}": field})

    return flat_schema, section_name_to_label


def handle_field_type(field: dict, field_values: np.array) -> Optional[Tuple[Variable, list]]:
    field_name = field['label']
    field_type = field['type']

    # Multiple choices fields are always discrete
    if 'choices' in field:
        discrete_values = sorted((choice['value'] for choice in field['choices']))
        var = DiscreteVariable(field_name, values=(str(val) for val in discrete_values))
        return var, [var.to_val(value) for value in field_values]
    else:
        discrete_values, _field_values, col_type = guess_data_type(field_values)
        if discrete_values is not None:
            # indexes = np.nonzero(_field_values[:, None] == discrete_values)[1]
            var = DiscreteVariable(field_name, values=(str(val) for val in discrete_values))
            return var, [var.to_val(str(value)) for value in _field_values]

        var = col_type(field_name)
        # Orange does not handle Time variables correctly
        if field_type in ('basic:date:', 'basic:datetime:'):
            _var = TimeVariable(field_name)
            _field_values = [_var.parse(time) for time in field_values]
            var = _var

        return var, _field_values


def descriptors(samples: List[Sample]) -> Dict[Variable, List[List[str]]]:
    def _drop_duplicate_descriptor_schemas(schemas: List[Dict[str, dict]]) -> List[Dict[str, dict]]:
        """
        Return a list where descriptor schemas are ensured to be unique.
        """
        return [json.loads(_schema) for _schema in {json.dumps(_schema) for _schema in schemas}]

    flatten_descriptor_schemas = [flatten_descriptor_schema(sample.descriptor_schema.schema) for sample in samples]
    schema, *tail = _drop_duplicate_descriptor_schemas([schema for schema, _ in flatten_descriptor_schemas])

    if len(tail):
        raise ValueError('Descriptor schema is not unique for selected samples.')

    section_name_to_label = [name_to_label for _, name_to_label in flatten_descriptor_schemas][0]
    meta_key_to_field = {key: field for key, field in schema.items()}
    meta_key_to_field.update(
        {
            'other.sample_name': {'label': 'Sample Name', 'type': 'basic:string:'},
            'other.sample_slug': {'label': 'Sample Slug', 'type': 'basic:string:'},
            'other.sample_id': {'label': 'Sample ID', 'type': 'basic:string:'},
        }
    )

    metadata = defaultdict(list)
    for sample in samples:
        descriptor = flatten_descriptor(sample.descriptor)

        for field_name in schema.keys():
            field = schema[field_name]
            descriptor_value = descriptor.get(field_name, '')

            if field['type'] == 'basic:boolean:' and descriptor_value is not None:
                descriptor_value = str(descriptor_value)

            metadata[field_name].append(descriptor_value)

        metadata['other.sample_name'].append(sample.name)
        metadata['other.sample_slug'].append(sample.slug)
        metadata['other.sample_id'].append(sample.id)

    # Filter fields with empty values
    metadata = {k: v for k, v in metadata.items() if not all(x == '' for x in v)}

    # handle duplicates
    split_section_field = [label.split(sep='.') for label in metadata.keys()]
    field_names = [field_name for _, field_name in split_section_field]
    sections = [section for section, _ in split_section_field]
    duplicates = {item for item, count in Counter(field_names).items() if count > 1}
    for key, section, field_name in zip(metadata.keys(), sections, field_names):
        if field_name in duplicates:
            meta_key_to_field[key]['label'] = f"{section_name_to_label[section]} {meta_key_to_field[key]['label']}"

    return dict(handle_field_type(meta_key_to_field[key], np.array(values)) for key, values in metadata.items())


def relations(samples: List[Sample], _relations: List[Relation]) -> Dict[Variable, List[List[str]]]:
    sample_ids = [sample.id for sample in samples]
    metadata = defaultdict(list)

    for relation in _relations:
        discrete_values = sorted({p['label'] for p in relation.partitions})
        sample_id_to_index = {p['entity']: discrete_values.index(p['label']) for p in relation.partitions}

        for samp_id in sample_ids:
            metadata[DiscreteVariable(relation.category, values=discrete_values)].append([sample_id_to_index[samp_id]])

    return metadata


def get_target_column(source: Table, incoming: Table) -> Variable:
    """
    Return matching column from both tables if there is any.
    """

    target_column_names = ('Sample ID', 'Sample slug', 'Sample name', 'Sample')

    column_candidates = [
        column
        for column in set(incoming.domain.metas) & set(source.domain.metas)
        if column.name in target_column_names
    ]

    if not column_candidates:
        raise ValueError(f'No matching columns detected. Use any of {target_column_names}')

    # Use the first column candidate. It does not really matter which one is used.
    return column_candidates[0]


def merge_clinical_data(source: Table, incoming: Table) -> Table:
    target_column = get_target_column(source, incoming)
    source_target_column, _ = source.get_column_view(source.domain.index(target_column))
    incoming_target_column, _ = incoming.get_column_view(incoming.domain.index(target_column))

    # We can not be sure that clinical data is in the same order (sample wise) as source data.
    # Map incoming row indexes to row indexes in source data
    # Example: X = ['B', 'C', 'A']
    #          Y = ['C', 'A', 'B']
    #          mask = [2, 0, 1] -> where should an element from array Y be to match array X
    source_index_by_row = {index: str(row_value) for index, row_value in enumerate(source_target_column)}
    incoming_row_by_index = {str(row_value): index for index, row_value in enumerate(incoming_target_column)}
    source_row_order = [
        incoming_row_by_index[source_index_by_row[row_idx]] for row_idx in range(len(source_target_column))
    ]

    # Columns from incoming table to include in source table
    incoming_columns = tuple(column for column in incoming.domain.metas if column not in source.domain.metas)

    # create new domain and transform data table
    domain = Domain(source.domain.variables, source.domain.class_vars, source.domain.metas + incoming_columns)
    # table = source.copy()
    table = source.transform(domain)

    for col in incoming_columns:
        col_view, _ = incoming.get_column_view(incoming.domain.index(col))
        table[:, col] = col_view[source_row_order].reshape(col_view.shape[0], 1)

    return table



# Test methods with long descriptive names can omit docstrings
# pylint: disable=missing-docstring,arguments-differ
import platform
import unittest
from itertools import chain
from unittest.mock import Mock, patch

import numpy as np

from Orange.data import Table, Domain
from Orange.projection import PCA
from Orange.data.filter import Values, FilterString
from Orange.widgets.tests.base import WidgetTest, WidgetOutputsTestMixin, ProjectionWidgetTestMixin, simulate
from Orange.widgets.unsupervised.owtsne import OWtSNE
from Orange.widgets.unsupervised.tests.test_owtsne import DummyTSNE, DummyTSNEModel

from orangecontrib.bioinformatics.utils import serverfiles
from orangecontrib.bioinformatics.widgets.utils.data import TAX_ID
from orangecontrib.bioinformatics.widgets.OWAnnotateProjection import OWAnnotateProjection

TIMEOUT = 20000


class TestOWAnnotateProjection(WidgetTest, ProjectionWidgetTestMixin, WidgetOutputsTestMixin):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        WidgetOutputsTestMixin.init(cls)
        cls._init_data()
        cls.signal_name = "Reference Data"
        cls.signal_data = cls.data
        cls.same_input_output_domain = False

        genes_path = serverfiles.localpath_download("marker_genes", "panglao_gene_markers.tab")
        filter_ = FilterString("Organism", FilterString.Equal, "Human")
        cls.genes = Values([filter_])(Table(genes_path))
        cls.genes.attributes[TAX_ID] = "9606"

    @classmethod
    def _init_data(cls):
        data_path = "https://datasets.biolab.si/sc/aml-1k.tab.gz"
        table_data = Table(data_path)
        table_data.attributes[TAX_ID] = "9606"

        ref_data = table_data[::2]
        pca = PCA(n_components=2)
        model = pca(ref_data)
        proj = model(ref_data)
        domain = Domain(
            ref_data.domain.attributes,
            ref_data.domain.class_vars,
            chain(ref_data.domain.metas, proj.domain.attributes),
        )
        cls.data = ref_data.transform(domain)
        cls.reference_data = ref_data
        cls.secondary_data = table_data[1:200:2]

    def setUp(self):
        self._patch_annotation_functions()
        self.widget = self.create_widget(OWAnnotateProjection)

    def _patch_annotation_functions(self):
        def mann_whitney_test(data):
            table = data.copy()
            table.X = np.random.random(data.X.shape)
            return table

        def assign_annotations(*args, **_):
            table = args[0].copy()
            table = table[:, 200]
            return table, table

        self.patcher1 = patch(
            "orangecontrib.bioinformatics.annotation." "annotate_samples.AnnotateSamples.mann_whitney_test",
            Mock(side_effect=mann_whitney_test),
        )
        self.patcher2 = patch(
            "orangecontrib.bioinformatics.annotation." "annotate_samples.AnnotateSamples.assign_annotations",
            Mock(side_effect=assign_annotations),
        )

        self.patcher1.start()
        self.patcher2.start()

    def tearDown(self):
        self.wait_until_finished(timeout=TIMEOUT)
        self.widget.cancel()
        self._restore_annotation_functions()

    def _restore_annotation_functions(self):
        self.patcher1.stop()
        self.patcher2.stop()

    def test_input_secondary_data(self):
        self.send_signal(self.widget.Inputs.secondary_data, self.secondary_data)
        self.wait_until_finished()
        self.assertTrue(self.widget.Error.no_reference_data.is_shown())

        self.send_signal(self.widget.Inputs.data, self.data)
        self.wait_until_finished()
        self.assertFalse(self.widget.Error.no_reference_data.is_shown())
        opts = self.widget.graph.ref_scatterplot_item.opts
        self.assertEqual(opts["pen"].color().name(), "#c8c8c8")
        self.assertEqual(opts["brush"].color().name(), "#c8c8c8")
        self.assertTrue(self.widget.graph.ref_scatterplot_item.isVisible())
        self.assertTrue(self.widget.graph.scatterplot_item.isVisible())

        self.send_signal(self.widget.Inputs.data, None)
        self.wait_until_finished()
        self.assertTrue(self.widget.Error.no_reference_data.is_shown())

        self.send_signal(self.widget.Inputs.secondary_data, None)
        self.wait_until_finished()
        self.assertFalse(self.widget.Error.no_reference_data.is_shown())

    def test_input_genes(self):
        self.send_signal(self.widget.Inputs.data, self.data)
        self.assertTrue(self.widget.Warning.no_genes.is_shown())
        self.wait_until_finished()
        self.send_signal(self.widget.Inputs.genes, self.genes)
        self.assertFalse(self.widget.Warning.no_genes.is_shown())
        self.wait_until_finished()
        self.send_signal(self.widget.Inputs.genes, None)
        self.assertTrue(self.widget.Warning.no_genes.is_shown())
        self.wait_until_finished()
        self.send_signal(self.widget.Inputs.data, None)
        self.assertFalse(self.widget.Warning.no_genes.is_shown())

    def test_scoring_method_control(self):
        self._restore_annotation_functions()
        self.send_signal(self.widget.Inputs.genes, self.genes)
        self.send_signal(self.widget.Inputs.data, self.data)
        self.wait_until_finished(timeout=TIMEOUT)
        output1 = self.get_output(self.widget.Outputs.annotated_data)
        cbox = self.widget.controls.scoring_method
        simulate.combobox_activate_index(cbox, 2)
        self.widget.run_button.click()
        self.wait_until_finished(timeout=TIMEOUT)
        output2 = self.get_output(self.widget.Outputs.annotated_data)
        np.testing.assert_array_equal(output1.X, output2.X)
        np.testing.assert_array_equal(output1.Y, output2.Y)
        self._patch_annotation_functions()

    def test_statistical_test_control(self):
        self._restore_annotation_functions()
        self.send_signal(self.widget.Inputs.genes, self.genes)
        self.send_signal(self.widget.Inputs.data, self.data)
        self.wait_until_finished(timeout=TIMEOUT)
        output1 = self.get_output(self.widget.Outputs.annotated_data)
        cbox = self.widget.controls.statistical_test
        simulate.combobox_activate_index(cbox, 1)
        self.widget.run_button.click()
        self.wait_until_finished(timeout=TIMEOUT)
        output2 = self.get_output(self.widget.Outputs.annotated_data)
        np.testing.assert_array_equal(output1.X, output2.X)
        np.testing.assert_array_equal(output1.Y, output2.Y)
        self.assertFalse(output1.metas.shape == output2.metas.shape and np.equal(output1.metas, output2.metas).all())
        self._patch_annotation_functions()

    def test_p_threshold_control(self):
        self._restore_annotation_functions()
        self.send_signal(self.widget.Inputs.genes, self.genes)
        self.send_signal(self.widget.Inputs.data, self.data)
        self.wait_until_finished(timeout=TIMEOUT)
        output1 = self.get_output(self.widget.Outputs.annotated_data)
        self.widget.controls.p_threshold.valueChanged.emit(0.1)
        self.widget.run_button.click()
        self.wait_until_finished(timeout=TIMEOUT)
        output2 = self.get_output(self.widget.Outputs.annotated_data)
        np.testing.assert_array_equal(output1.X, output2.X)
        np.testing.assert_array_equal(output1.Y, output2.Y)
        self._patch_annotation_functions()

    def test_epsilon_control(self):
        self.send_signal(self.widget.Inputs.data, self.data)
        self.wait_until_finished()
        self.assertFalse(self.widget.epsilon_spin.isEnabled())
        self.widget.controls.use_user_epsilon.click()
        self.assertTrue(self.widget.epsilon_spin.isEnabled())
        self.assertFalse(self.widget.Information.modified.is_shown())
        self.widget.controls.use_user_epsilon.click()
        self.assertTrue(self.widget.Information.modified.is_shown())

    def test_output_data(self):
        self._restore_annotation_functions()
        self.send_signal(self.widget.Inputs.genes, self.genes)
        self.send_signal(self.widget.Inputs.data, self.data)
        self.wait_until_finished(timeout=TIMEOUT)
        output = self.get_output(self.widget.Outputs.annotated_data)
        n_metas = len(self.data.domain.metas)
        self.assertGreater(len(output.domain.metas), n_metas + 4)
        self.assertListEqual(
            ["Clusters", "Annotation"] + [m.name for m in self.data.domain.metas] + ["Selected"],
            [m.name for m in output.domain.metas[-n_metas - 3 :]],
        )
        np.testing.assert_array_equal(output.X, self.data.X)
        np.testing.assert_array_equal(output.Y, self.data.Y)
        np.testing.assert_array_equal(output.metas[:, -n_metas - 1 : -1], self.data.metas)
        self._patch_annotation_functions()

    def test_button_no_data(self):
        self.widget.run_button.click()
        self.assertEqual(self.widget.run_button.text(), "Start")

    def test_button_with_data(self):
        self.send_signal(self.widget.Inputs.genes, self.genes)
        self.send_signal(self.widget.Inputs.data, self.data)
        self.assertEqual(self.widget.run_button.text(), "Stop")
        self.wait_until_finished()
        self.assertEqual(self.widget.run_button.text(), "Start")

    def test_button_toggle(self):
        self.send_signal(self.widget.Inputs.genes, self.genes)
        self.send_signal(self.widget.Inputs.data, self.data)
        self.widget.run_button.click()
        self.assertEqual(self.widget.run_button.text(), "Resume")

    def test_plot_once(self):
        self.widget.setup_plot = Mock()
        self.widget.send_data = Mock()
        self.send_signal(self.widget.Inputs.genes, self.genes)
        self.widget.setup_plot.assert_called_once()
        self.widget.setup_plot.reset_mock()
        self.widget.send_data.assert_called_once()
        self.widget.send_data.reset_mock()
        self.wait_until_finished()
        self.send_signal(self.widget.Inputs.data, self.data)
        self.widget.setup_plot.assert_called_once()
        self.widget.send_data.assert_called_once()
        self.wait_until_finished()
        self.widget.setup_plot.reset_mock()
        self.widget.send_data.reset_mock()
        self.wait_until_finished()
        self.send_signal(self.widget.Inputs.secondary_data, self.secondary_data)
        self.widget.setup_plot.assert_called_once()
        self.widget.send_data.assert_called_once()

    def test_saved_selection(self):
        self.send_signal(self.widget.Inputs.genes, self.genes)
        self.send_signal(self.widget.Inputs.data, self.data)
        self.wait_until_finished()

        self.widget.graph.select_by_indices(list(range(0, len(self.data), 10)))
        settings = self.widget.settingsHandler.pack_data(self.widget)
        widget = self.create_widget(self.widget.__class__, stored_settings=settings)

        self.send_signal(widget.Inputs.genes, self.genes, widget=widget)
        self.send_signal(widget.Inputs.data, self.data, widget=widget)
        self.wait_until_finished(widget=widget)

        self.assertEqual(np.sum(widget.graph.selection), 50)
        np.testing.assert_equal(self.widget.graph.selection, widget.graph.selection)

    def test_outputs(self, timeout=TIMEOUT):
        self._restore_annotation_functions()
        self.send_signal(self.widget.Inputs.genes, self.genes)
        super().test_outputs(timeout=timeout)
        self._patch_annotation_functions()

    def test_color_by_cluster(self):
        self._restore_annotation_functions()
        self.send_signal(self.widget.Inputs.genes, self.genes)
        self.send_signal(self.widget.Inputs.data, self.data)
        self.wait_until_finished(timeout=TIMEOUT)
        self.assertTrue(self.widget.controls.attr_color.isEnabled())

        simulate.combobox_activate_index(self.widget.controls.attr_color, 0)
        is_grey = [brush.color().name() == "#808080" for brush in self.widget.graph.scatterplot_item.data['brush']]
        self.assertTrue(all(is_grey))
        self.widget.controls.color_by_cluster.click()
        self.assertFalse(self.widget.controls.attr_color.isEnabled())
        is_not_grey = [brush.color().name() != "#808080" for brush in self.widget.graph.scatterplot_item.data['brush']]
        self.assertTrue(any(is_not_grey))
        self._patch_annotation_functions()

    def test_attr_label_metas(self):
        self.send_signal(self.widget.Inputs.data, self.data)
        self.wait_until_finished()
        simulate.combobox_activate_item(self.widget.controls.attr_label, self.data.domain[-1].name)

    def test_attr_models(self):
        self.send_signal(self.widget.Inputs.data, self.data)
        self.wait_until_finished()
        controls = self.widget.controls
        self.assertEqual(len(controls.attr_color.model()), 1008)
        self.assertEqual(len(controls.attr_shape.model()), 5)
        self.assertEqual(len(controls.attr_size.model()), 1005)
        self.assertEqual(len(controls.attr_label.model()), 1010)
        for var in self.data.domain.variables + self.data.domain.metas:
            self.assertIn(var, controls.attr_label.model())
            if var.is_continuous:
                self.assertIn(var, controls.attr_color.model())
                self.assertIn(var, controls.attr_size.model())
                self.assertNotIn(var, controls.attr_shape.model())
            if var.is_discrete:
                self.assertIn(var, controls.attr_color.model())
                self.assertNotIn(var, controls.attr_size.model())
                self.assertIn(var, controls.attr_shape.model())

    def test_subset_data_color(self):
        self.assertRaises(AttributeError, lambda: self.widget.Inputs.data_subset)

    def test_sparse_data(self):
        table = Table("iris").to_sparse()
        self.send_signal(self.widget.Inputs.data, table)

    def test_missing_embedding_data(self):
        self.send_signal(self.widget.Inputs.genes, self.genes)
        self.send_signal(self.widget.Inputs.data, self.secondary_data[::2])
        self.wait_until_finished()
        secondary_data = self.secondary_data[1::2]
        self.send_signal(self.widget.Inputs.secondary_data, secondary_data)
        self.wait_until_finished()
        self.assertFalse(self.widget.Warning.missing_compute_value.is_shown())
        secondary_data = secondary_data[:, 2:]
        self.send_signal(self.widget.Inputs.secondary_data, secondary_data)
        self.wait_until_finished()
        self.assertTrue(self.widget.Warning.missing_compute_value.is_shown())
        self.send_signal(self.widget.Inputs.secondary_data, None)
        self.wait_until_finished()
        self.assertFalse(self.widget.Warning.missing_compute_value.is_shown())

    @patch("Orange.projection.manifold.TSNE", DummyTSNE)
    @patch("Orange.projection.manifold.TSNEModel", DummyTSNEModel)
    def test_tsne_output(self):
        owtsne = self.create_widget(OWtSNE)
        self.send_signal(owtsne.Inputs.data, self.reference_data, widget=owtsne)
        self.wait_until_finished(widget=owtsne, timeout=TIMEOUT)
        tsne_output = self.get_output(owtsne.Outputs.annotated_data, owtsne)

        self.send_signal(self.widget.Inputs.genes, self.genes)
        self.send_signal(self.widget.Inputs.data, tsne_output)
        self.wait_until_finished()
        self.send_signal(self.widget.Inputs.secondary_data, self.secondary_data)

    @unittest.skipIf(platform.system() == 'Windows', 'Windows fatal exception: access violation')
    def test_send_report(self):
        super().test_send_report()


if __name__ == "__main__":
    unittest.main()



""" ResolweAPI """
from typing import Dict, List
from pathlib import Path
from tempfile import NamedTemporaryFile
from urllib.parse import urljoin

import requests
from resdk import Resolwe
from requests import Response
from resdk.resources.data import Data, Sample

from Orange.data import Table, Variable

from orangecontrib.bioinformatics.resolwe import sample_metadata

DEFAULT_URL: str = 'https://app.genialis.com'
RESOLWE_URLS: List[str] = [DEFAULT_URL, 'https://imaps.genialis.com', 'https://bcm.genialis.com']
CREDENTIAL_MANAGER_SERVICE: str = 'resolwe_credentials'


class ResolweAPI:

    COLLECTION_FIELDS = (
        'id',
        'slug',
        'name',
        'created',
        'modified',
        'description',
        'data_count',
        'entity_count',
        'contributor__first_name',
        'contributor__last_name',
        'contributor__username',
        'tags',
    )

    def __init__(self, username=None, password=None, url=DEFAULT_URL):
        self._res = Resolwe(username, password, url)

    @property
    def url(self):
        return self._res.url

    @property
    def base_url(self) -> str:
        return self._res.api._store['base_url']

    @property
    def collection_url(self):
        return f'{self.url}/expressions/search/collection/'

    @property
    def session(self) -> requests.Session:
        return self._res.session

    def get_currently_logged_user(self):
        url = urljoin(self.base_url, 'user?current_only=1')
        response = self.session.get(url, auth=self._res.auth)
        response.raise_for_status()
        return response.json()

    def get_collections(self, **kwargs):
        kwargs.update({'fields': ','.join(self.COLLECTION_FIELDS)})
        params = '&'.join([f'{key}={value}' for key, value in kwargs.items() if value is not None])
        url = urljoin(self.base_url, f'collection?{params}')

        response = self.session.get(url, auth=self._res.auth)
        response.raise_for_status()
        return response.json()

    def get_species(self, collection_ids: List[str]) -> Dict[str, str]:
        params = '&'.join(f'collection[]={col_id}' for col_id in collection_ids)
        url = urljoin(self.base_url, f'_modules/species/collection?{params}')

        response = self.session.get(url, auth=self._res.auth)
        response.raise_for_status()
        response_data = response.json()

        col_to_species = {
            data['collection_id']: data['species'][0] if len(data['species']) == 1 else '' for data in response_data
        }

        return col_to_species

    def download(self, data_id, data_file_name) -> Response:
        file_url = urljoin(self.url, f'data/{data_id}/{data_file_name}')
        response = self.session.get(file_url, auth=self._res.auth)
        response.raise_for_status()
        return response

    def get_expression_data_objects(self, collection_id: str) -> List[Data]:
        """ Get all expression data objects from a collection. """
        return self._res.data.filter(type='data:expression', collection=collection_id)

    def get_orange_data_objects(self, collection_id: str) -> List[Data]:
        return self._res.data.filter(type='data:metadata:orange', collection=collection_id, ordering='-modified')

    def get_sample_relations(self, rel_type: str, collection_id: str):
        """ Get all relations (by type) from a collection. """
        return self._res.relation.filter(collection=collection_id).filter(type=rel_type)

    def get_samples_by_id(self, sample_ids: List[str]) -> List[Sample]:
        return self._res.sample.filter(id__in=sample_ids)

    def aggregate_sample_metadata(self, samples: List[Sample], collection_id: str) -> Dict[Variable, List[List[str]]]:
        replicate_groups = list(self.get_sample_relations('group', collection_id))
        replicate_series = list(self.get_sample_relations('series', collection_id))

        descriptors = sample_metadata.descriptors(samples)
        relations = sample_metadata.relations(samples, replicate_groups + replicate_series)

        metadata = {}
        metadata.update(descriptors)
        metadata.update(relations)

        return metadata

    def get_clinical_metadata(self, data_id, data_file_name) -> Table:
        response = self.download(data_id, data_file_name)
        with NamedTemporaryFile(suffix=''.join(Path(data_file_name).suffixes)) as f:
            f.write(response.content)
            f.seek(0)
            return Table.from_file(f.name)


if __name__ == '__main__':
    res = ResolweAPI(url=DEFAULT_URL)
    print(res)



from typing import Dict, List, Optional
from functools import reduce
from collections import defaultdict

from orangecontrib.bioinformatics.utils import serverfiles
from orangecontrib.bioinformatics.ncbi.gene import Gene


def _from_data_to_gene(line: str):
    gene = Gene()
    for attr, val in zip(['homology_group_id', 'tax_id', 'gene_id'], line.strip().split('\t')):
        setattr(gene, attr, val)
    return gene


class HomoloGene:
    """ Wrapper around NCBI HomoloGene database """

    def __init__(self):
        self.file_path: str = serverfiles.localpath_download('homologene', 'homologene.tab')

        with open(self.file_path, 'r') as fp:
            self._homologs: Dict[str, Gene] = {
                h.gene_id: h for h in [_from_data_to_gene(line) for line in fp.readlines()]
            }

        def _helper(groups, gene):
            groups[gene.homology_group_id].append(gene)
            return groups

        self._homologs_by_group: Dict[str, List[Gene]] = reduce(_helper, self._homologs.values(), defaultdict(list))

    def find_homolog(self, gene_id: str, organism: str) -> Optional[str]:
        """ Find homolog gene in organism. If the homolog does not exist, return None. """
        homology_group = self._homologs.get(str(gene_id), Gene()).homology_group_id
        homologs = [
            gene.gene_id for gene in self._homologs_by_group.get(homology_group, []) if gene.tax_id == organism
        ]
        if len(homologs) == 1:
            return homologs[0]
        else:
            # Is possible that find more then one gene?
            return None


if __name__ == "__main__":
    from orangecontrib.bioinformatics.ncbi.gene import GeneMatcher, load_gene_summary
    import Orange

    homology = HomoloGene()

    gm = GeneMatcher('4932')
    genes = Orange.data.Table("brown-selected")

    gm.genes = genes
    _homologs = [homology.find_homolog(str(gene.gene_id), '9606') for gene in gm.genes]
    _homologs = load_gene_summary('9606', _homologs)

    for gene, homolog in zip(gm.genes, _homologs):
        print(f'{gene} ----> {homolog}')



import warnings
from functools import singledispatch
from itertools import chain, starmap
from collections import namedtuple, defaultdict
from xml.sax.saxutils import escape

import numpy as np

from AnyQt.QtGui import QStandardItem, QStandardItemModel
from AnyQt.QtCore import Qt, QSize, QItemSelection, QItemSelectionModel, QPersistentModelIndex
from AnyQt.QtCore import pyqtSignal as Signal
from AnyQt.QtWidgets import QWidget, QComboBox, QGroupBox, QListView, QSizePolicy, QVBoxLayout

from Orange.data import Variable, StringVariable, DiscreteVariable, ContinuousVariable
from Orange.widgets import gui
from Orange.widgets.utils import itemmodels


@singledispatch
def standarditem_from(obj):
    item = QStandardItem(str(obj))
    item.setData(obj, Qt.UserRole)
    item.setFlags(item.flags() & ~Qt.ItemIsEditable)
    return item


@standarditem_from.register(Variable)
def standarditem_from_var(var):
    item = QStandardItem(var.name)
    _, icon = gui.attributeItem(var)
    item.setIcon(icon)
    item.setToolTip(itemmodels.Variable)


def variable_tooltip(var):
    if isinstance(var, DiscreteVariable):
        return discrete_variable_tooltip(var)
    elif isinstance(var, ContinuousVariable):
        return continuous_variable_toltip(var)
    elif isinstance(var, StringVariable):
        return string_variable_tooltip(var)


def variable_labels_tooltip(var):
    text = ""
    if var.attributes:
        items = [(escape(key), escape(value)) for key, value in var.attributes.items()]
        labels = list(starmap("{!s} = {!s}".format, items))
        text += "<br/>Variable Labels:<br/>"
        text += "<br/>".join(labels)
    return text


def discrete_variable_tooltip(var):
    text = "<b>%s</b><br/>Discrete with %i values: " % (escape(var.name), len(var.values))
    text += ", ".join("%r" % escape(v) for v in var.values)
    text += variable_labels_tooltip(var)
    return text


def continuous_variable_toltip(var):
    text = "<b>%s</b><br/>Continuous" % escape(var.name)
    text += variable_labels_tooltip(var)
    return text


def string_variable_tooltip(var):
    text = "<b>%s</b><br/>String" % escape(var.name)
    text += variable_labels_tooltip(var)
    return text


def itemselection(modelindexlist):
    """
    Return an QtCore.QItemSelection from QModelIndex list

    Parameters
    ----------
    modelindexlist : list of QtCore.QModelIndex
        Selected model indices.

    Returns
    -------
    selection : QtCore.QItemSelection
    """
    selection = QItemSelection()
    for index in modelindexlist:
        selection.select(index, index)
    return selection


ColumnGroup = namedtuple("ColumnGroup", ["name", "key", "values"])
RowGroup = namedtuple("RowGroup", ["name", "var", "values"])


def group_candidates(data):
    items = [attr.attributes.items() for attr in data.domain.attributes]
    items = list(chain(*items))

    targets = defaultdict(set)
    for label, value in items:
        targets[label].add(value)

    # Need at least 2 distinct values or key
    targets = [(key, sorted(vals)) for key, vals in targets.items() if len(vals) >= 2]

    column_groups = [ColumnGroup(key, key, values) for key, values in sorted(targets)]

    disc_vars = [
        var
        for var in data.domain.class_vars + data.domain.metas
        if isinstance(var, DiscreteVariable) and len(var.values) >= 2
    ]

    row_groups = [RowGroup(var.name, var, var.values) for var in disc_vars]
    return column_groups, row_groups


@standarditem_from.register(ColumnGroup)
def standarditem_from_columngroup(colgroup):
    item = QStandardItem(colgroup.name)
    item.setToolTip("Split by column label: '{!s}'".format(escape(colgroup.name)))
    item.setFlags(item.flags() & ~Qt.ItemIsEditable)
    item.setData(colgroup, Qt.UserRole)
    children = [standarditem_from(val) for val in colgroup.values]
    item.appendRows(children)
    return item


@standarditem_from.register(RowGroup)
def standarditem_from_rowgroup(rowgroup):
    item = QStandardItem(rowgroup.name)
    icon, _ = gui.attributeItem(rowgroup.var)
    item.setIcon(icon)
    item.setToolTip(variable_tooltip(rowgroup.var))
    item.setData(rowgroup, Qt.UserRole)
    item.setFlags(item.flags() & ~Qt.ItemIsEditable)
    children = [standarditem_from(val) for val in rowgroup.values]
    item.appendRows(children)
    return item


def group_selection_mask(data, group, indices):
    """ Return the selection masks for the group.
    """
    if isinstance(group, ColumnGroup):
        selected = [group.values[i] for i in indices]
        target = {(group.key, value) for value in selected}
        _i = [bool(set(var.attributes.items()).intersection(target)) for var in data.domain.attributes]
        return np.array(_i, dtype=bool)
    elif isinstance(group, RowGroup):
        target = set(indices)
        x, _ = data.get_column_view(group.var)
        _i = np.zeros_like(x, dtype=bool)
        for i in target:
            _i |= x == i
        return _i
    else:
        raise TypeError("ColumnGroup or RowGroup expected, got {}".format(type(group).__name__))


class LabelSelectionWidget(QWidget):
    """
    A widget for selection of label values.

    The widget displays the contents of the model with two widgets:

    * The top level model items are displayed in a combo box.
    * The children of the current selected top level item are
      displayed in a subordinate list view.

    .. note:: This is not a QAbstractItemView subclass.
    """

    # Current group/root index has changed.
    groupChanged = Signal(int)
    # Selection for the current group/root has changed.
    groupSelectionChanged = Signal(int)

    def __init__(self):
        super().__init__()
        self.__model = None
        self.__selectionMode = QListView.ExtendedSelection

        self.__currentIndex = -1
        self.__selections = {}

        self.__parent = None

        self.targets = []

        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)

        def group_box(title):
            box = QGroupBox(title)
            box.setFlat(True)
            lay = QVBoxLayout()
            lay.setContentsMargins(0, 0, 0, 0)
            box.setLayout(lay)
            return box

        self.labels_combo = QComboBox()
        self.values_view = QListView(selectionMode=self.__selectionMode)

        self.labels_combo.currentIndexChanged.connect(self.__onCurrentIndexChanged)

        l_box = group_box(self.tr("Label"))
        v_box = group_box(self.tr("Values"))

        l_box.layout().addWidget(self.labels_combo)
        v_box.layout().addWidget(self.values_view)

        layout.addWidget(l_box)
        layout.addWidget(v_box)

        self.setLayout(layout)

        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

    def set_selection(self):
        model = self.model()

        # FIX: This assumes fixed target key/values order.
        indices = [
            model.index(i, 0, model.index(keyind, 0))
            for keyind, selected in enumerate(self.__parent.stored_selections)
            for i in selected
        ]

        selection = itemselection(indices)
        self.setCurrentGroupIndex(self.__parent.current_group_index)
        self.setSelection(selection)

    def selected_split(self):
        group_index = self.currentGroupIndex()
        if not (0 <= group_index < len(self.targets)):
            return None, []

        group = self.targets[group_index]
        selection = [ind.row() for ind in self.currentGroupSelection().indexes()]
        return group, selection

    def set_data(self, parent, data):
        """ Initialize widget state after receiving new data.
        """
        self.__parent = parent
        # self.__currentIndex = parent.current_group_index
        if data is not None:
            column_groups, row_groups = group_candidates(data)

            self.targets = column_groups + row_groups

            model = QStandardItemModel()
            for item in [standarditem_from(desc) for desc in self.targets]:
                model.appendRow(item)

            self.setModel(model)
        else:
            self.targets = []
            self.setModel(None)

    def clear(self):
        """ Clear the widget/model (same as ``setModel(None)``).
        """
        if self.__model is not None:
            self.values_view.selectionModel().clearSelection()
            self.values_view.selectionModel().selectionChanged.disconnect(self.__onSelectionChanged)

            self.values_view.setModel(None)
            self.labels_combo.setModel(QStandardItemModel(self.labels_combo))
            self.__currentIndex = -1
            self.__selections = {}
            self.__model = None

    def setModel(self, model):
        """
        Set the source model for display.

        The model should be a tree model with depth 2.

        Parameters
        ----------
        model : QtCore.QAbstractItemModel
        """
        if model is self.__model:
            return

        self.clear()

        if model is None:
            return

        self.__model = model
        self.values_view.setModel(model)
        self.values_view.setRootIndex(model.index(0, 0))

        self.values_view.selectionModel().selectionChanged.connect(self.__onSelectionChanged)
        # will emit the currentIndexChanged (if the model is not empty)
        self.labels_combo.setModel(model)

    def model(self):
        """
        Return the current model.

        Returns
        -------
        model : QtCore.QAbstractItemModel
        """
        return self.__model

    def setCurrentGroupIndex(self, index):
        """
        Set the current selected group/root index row.

        Parameters
        ----------
        index : int
            Group index.
        """
        self.labels_combo.setCurrentIndex(index)

    def currentGroupIndex(self):
        """
        Return the current selected group/root index row.

        Returns
        -------
        row : index
            Current group row index (-1 if there is no current index)
        """
        return self.labels_combo.currentIndex()

    def setSelection(self, selection):
        """
        Set the model item selection.

        Parameters
        ----------
        selection : QtCore.QItemSelection
            Item selection.
        """
        if self.values_view.selectionModel() is not None:
            indices = selection.indexes()
            pind = defaultdict(list)

            for index in indices:
                parent = index.parent()
                if parent.isValid():
                    if parent == self.__model.index(parent.row(), parent.column()):
                        pind[parent.row()].append(QPersistentModelIndex(index))
                    else:
                        warnings.warn("Die Die Die")
                else:
                    # top level index
                    pass

            self.__selections = pind
            self.__restoreSelection()

    def selection(self):
        """
        Return the item selection.

        Returns
        -------
        selection : QtCore.QItemSelection
        """
        selection = QItemSelection()
        if self.__model is None:
            return selection

        for pind in chain(*self.__selections.values()):
            ind = self.__model.index(pind.row(), pind.column(), pind.parent())
            if ind.isValid():
                selection.select(ind, ind)
        return selection

    def currentGroupSelection(self):
        """
        Return the item selection for the current group only.
        """
        if self.values_view.selectionModel() is not None:
            return self.values_view.selectionModel().selection()
        else:
            return QItemSelection()

    def __onCurrentIndexChanged(self, index):
        self.__storeSelection(self.__currentIndex, self.values_view.selectedIndexes())

        self.__currentIndex = index
        if self.__model is not None:
            root = self.__model.index(index, 0)
            self.values_view.setRootIndex(root)

            self.__restoreSelection()
        self.groupChanged.emit(index)

    def __onSelectionChanged(self, old, new):
        self.__storeSelection(self.__currentIndex, self.values_view.selectedIndexes())

        self.groupSelectionChanged.emit(self.__currentIndex)

    def __storeSelection(self, groupind, indices):
        # Store current values selection for the current group
        groupind = self.__currentIndex
        indices = [QPersistentModelIndex(ind) for ind in self.values_view.selectedIndexes()]
        self.__selections[groupind] = indices

    def __restoreSelection(self):
        # Restore previous selection for root (if available)
        assert self.__model is not None
        groupind = self.__currentIndex
        root = self.__model.index(groupind, 0)
        sel = self.__selections.get(groupind, [])
        indices = [
            self.__model.index(pind.row(), pind.column(), root)
            for pind in sel
            if pind.isValid() and pind.parent() == root
        ]

        selection = QItemSelection()
        for ind in indices:
            selection.select(ind, ind)
        self.values_view.selectionModel().select(selection, QItemSelectionModel.ClearAndSelect)

    def sizeHint(self):
        """Reimplemented from QWidget.sizeHint"""
        return QSize(100, 200)



""" dictyExpress widget """
import sys
import threading

from requests.exceptions import ConnectionError

from AnyQt.QtGui import QFont
from AnyQt.QtCore import Qt, QSize, QThreadPool
from AnyQt.QtWidgets import QFrame, QLabel, QLineEdit, QTreeWidget, QApplication, QTreeWidgetItem

from Orange.data import Table, StringVariable
from Orange.widgets import gui, settings
from Orange.widgets.widget import Msg, OWWidget
from Orange.widgets.credentials import CredentialManager
from Orange.widgets.utils.signals import Output
from Orange.widgets.utils.concurrent import ConcurrentWidgetMixin

from orangecontrib.bioinformatics import resolwe
from orangecontrib.bioinformatics.ncbi.gene import ENTREZ_ID, GeneMatcher
from orangecontrib.bioinformatics.resolwe.utils import etc_to_table
from orangecontrib.bioinformatics.widgets.utils.data import (
    TAX_ID,
    GENE_ID_COLUMN,
    GENE_ID_ATTRIBUTE,
    GENE_AS_ATTRIBUTE_NAME,
)

Labels = [
    (" ", " "),
    ("var.project", "Project"),
    ("static.name", "Experiment"),
    ("static.cite", "Citation"),
    ("var.growth", "Growth"),
    ("var.treatment", "Treatment"),
    ("var.strain", "Strain"),
]


# Creates line separator
def h_line():
    line = QFrame()
    line.setFrameShape(QFrame.HLine)
    line.setFrameShadow(QFrame.Sunken)
    return line


class OWdictyExpress(OWWidget, ConcurrentWidgetMixin):

    name = "dictyExpress"
    description = "Time-course gene expression data"
    icon = "../widgets/icons/OWdictyExpress.svg"
    want_main_area = True
    priority = 20

    class Inputs:
        pass

    class Outputs:
        etc_data = Output("Data", Table)

    class Error(OWWidget.Error):
        unreachable_host = Msg('Host not reachable')
        invalid_credentials = Msg('Invalid credentials')

    gene_as_attr_name = settings.Setting(0)

    selected_item = settings.Setting(None, schema_only=True)
    auto_commit = settings.Setting(False, schema_only=True)

    def __init__(self):
        super().__init__()
        ConcurrentWidgetMixin.__init__(self)

        self.res = None
        self.organism = '44689'
        self.server = 'https://dictyexpress.research.bcm.edu'
        self.headerLabels = [x[1] for x in Labels]
        self.searchString = ""
        self.items = []
        self.cm: CredentialManager = CredentialManager('genapi_credentials')

        self.progress_bar = None
        # threads
        self.threadpool = QThreadPool()

        # Login Section
        box = gui.widgetBox(self.controlArea, 'Login')

        self.username_line_edit = gui.lineEdit(
            box, self, "", "Username:", labelWidth=100, orientation='horizontal', callback=self.handle_sign_in_btn
        )

        self.password_line_edit = gui.lineEdit(
            box, self, "", "Password:", labelWidth=100, orientation='horizontal', callback=self.handle_sign_in_btn
        )
        self.password_line_edit.setEchoMode(QLineEdit.Password)

        box = gui.widgetBox(box, orientation=Qt.Horizontal)
        self.sign_in_btn = gui.button(box, self, 'Sign In', autoDefault=False, callback=self.sign_in)
        self.sign_out_btn = gui.button(box, self, 'Sign Out', autoDefault=False, callback=self.sign_out)
        self.sign_in_btn.setDisabled(True)

        self.controlArea.layout().addWidget(h_line())

        box = gui.widgetBox(self.controlArea, "Output", addSpace=True)
        gui.radioButtonsInBox(
            box, self, "gene_as_attr_name", ["Genes in rows", "Genes in columns"], callback=self.invalidate
        )

        self.controlArea.layout().addWidget(h_line())

        self.clear_cache_btn = gui.button(
            self.controlArea, self, "Clear cache", autoDefault=False, callback=self.clear_cache
        )

        gui.rubber(self.controlArea)

        self.commit_button = gui.auto_commit(self.controlArea, self, "auto_commit", "&Commit", box=False)

        # Experiment Section

        label = QLabel("Available projects:")
        my_font = QFont()
        my_font.setBold(True)
        label.setFont(my_font)
        self.mainArea.layout().addWidget(label)

        self.mainArea.layout().addWidget(h_line())

        self.filter = gui.lineEdit(
            self.mainArea, self, "searchString", "Filter:", callbackOnType=True, callback=self.search_update
        )

        self.experimentsWidget = QTreeWidget(
            alternatingRowColors=True, rootIsDecorated=False, uniformRowHeights=True, sortingEnabled=True
        )

        self.experimentsWidget.setItemDelegateForColumn(0, gui.IndicatorItemDelegate(self, role=Qt.DisplayRole))

        self.experimentsWidget.selectionModel().selectionChanged.connect(self.on_selection_changed)

        self.experimentsWidget.setHeaderLabels(self.headerLabels)
        self.mainArea.layout().addWidget(self.experimentsWidget)

        self.sign_in()
        self.sizeHint()

    def sizeHint(self):
        return QSize(1400, 680)

    def handle_sign_in_btn(self):
        self.sign_in_btn.setEnabled(
            True if self.username_line_edit.text() and self.password_line_edit.text() else False
        )

    def sign_out(self):
        self.username_line_edit.setText(None)
        self.password_line_edit.setText(None)
        del self.cm.username
        del self.cm.password
        self.reset()
        self.sign_in()

    def sign_in(self):
        self.res = None
        self.Error.clear()
        self.reset()

        username = self.cm.username if self.cm.username else self.username_line_edit.text()
        password = self.cm.password if self.cm.password else self.password_line_edit.text()

        try:
            self.res = resolwe.connect(
                username if username else resolwe.DEFAULT_EMAIL,
                password if password else resolwe.DEFAULT_PASSWD,
                self.server,
                'genesis',
            )

        except resolwe.ResolweAuthException:
            self.Error.invalid_credentials()
        else:
            self.load_experiments()
            # self.handle_cache_button(True)

            self.cm.username = username
            self.cm.password = password
            self.username_line_edit.setText(username)
            self.password_line_edit.setText(password)

    def clear_cache(self):
        resolwe.GenAPI.clear_cache()
        self.reset()
        self.load_experiments()

    def reset(self):
        self.experimentsWidget.clear()  # clear QTreeWidget
        self.items = []
        # self.lastSelected = None
        self.searchString = ""

    def search_update(self):
        parts = self.searchString.split()
        for item in self.items:
            item.setHidden(not all(s in item for s in parts))

    def progress_advance(self):
        # GUI should be updated in main thread. That's why we are calling advance method here
        assert threading.current_thread() == threading.main_thread()
        if self.progress_bar:
            self.progress_bar.advance()

    def load_experiments_result(self, experiments):
        self.load_tree_items(experiments)
        self.progress_bar.finish()
        self.setStatusMessage('')

    def on_exception(self, ex):
        if isinstance(ex, ConnectionError) or isinstance(ex, ValueError):
            self.Error.unreachable_host()

        print(ex)

    def on_done(self, results):
        if isinstance(results, list):
            self.load_tree_items(results)
        elif isinstance(results, tuple):
            self.send_to_output(results)

    def load_experiments(self):
        if self.res:
            self.start(self.res.fetch_etc_objects)

    def load_tree_items(self, list_of_exp):
        self.items = [CustomTreeItem(self.experimentsWidget, item) for item in list_of_exp]

        for i in range(len(self.headerLabels)):
            self.experimentsWidget.resizeColumnToContents(i)

        self.set_cached_indicator()
        self.set_selected()

    def set_selected(self):
        for item in self.items:
            if self.selected_item and item.gen_data_id == self.selected_item:
                self.experimentsWidget.setCurrentItem(item)

    def on_selection_changed(self):
        self.invalidate()

    def invalidate(self):
        self.commit()

    def send_to_output(self, result):
        etc_json, table_name = result

        # convert to table
        data = etc_to_table(etc_json, bool(self.gene_as_attr_name))
        # set table name
        data.name = table_name

        # match genes
        gene_matcher = GeneMatcher(str(self.organism))

        if not bool(self.gene_as_attr_name):
            if 'Gene' in data.domain:
                data = gene_matcher.match_table_column(data, 'Gene', StringVariable(ENTREZ_ID))
            data.attributes[GENE_ID_COLUMN] = ENTREZ_ID
        else:
            data = gene_matcher.match_table_attributes(data)
            data.attributes[GENE_ID_ATTRIBUTE] = ENTREZ_ID

        # add table attributes
        data.attributes[TAX_ID] = str(self.organism)
        data.attributes[GENE_AS_ATTRIBUTE_NAME] = bool(self.gene_as_attr_name)

        # reset cache indicators
        self.set_cached_indicator()
        # send data to the output signal
        self.Outputs.etc_data.send(data)

    def commit(self):
        self.Error.clear()
        selected_items = self.experimentsWidget.selectedItems()  # get selected TreeItem

        if len(selected_items) < 1:
            self.Outputs.etc_data.send(None)
            return

        selected_item = selected_items[0]
        self.selected_item = selected_item.gen_data_id
        self.start(self.res.download_etc_data, selected_item.gen_data_id, table_name=selected_item.data_name)

    def set_cached_indicator(self):
        cached = self.res.get_cached_ids()

        for item in self.items:

            if item.gen_data_id in cached:
                item.setData(0, Qt.DisplayRole, " ")
            else:
                item.setData(0, Qt.DisplayRole, "")


class CustomTreeItem(QTreeWidgetItem):
    def __init__(self, parent, gen_data):
        super(CustomTreeItem, self).__init__(parent)  # Init super class (QtGui.QTreeWidgetItem )

        self._gen_data = gen_data  # GenData object
        self.set_rows(self._gen_data.annotation)  # set rows in QTreeWidget

    def __contains__(self, text):
        return any(text.upper() in str(self.text(i)).upper() for i in range(self.columnCount()))

    @property
    def gen_data_id(self):
        return self._gen_data.id

    @property
    def data_name(self):
        try:
            project = self._gen_data.var['project']
            experiment = self._gen_data.static['name']
        except (AttributeError, KeyError):
            project = ''
            experiment = ''

        return '{} ({})'.format(project, experiment)

    def set_rows(self, row):
        for index, label in enumerate(Labels):
            if index > 0:
                try:
                    if type(row[label[0]]["value"]) == list:
                        self.setText(index, row[label[0]]["value"][0]["name"])
                    else:
                        self.setText(index, row[label[0]]["value"])
                except (IndexError, KeyError):
                    self.setText(index, 'No data')


if __name__ == "__main__":

    def test_main():
        app = QApplication(sys.argv)
        w = OWdictyExpress()
        w.show()
        r = app.exec_()
        w.saveSettings()
        return r

    sys.exit(test_main())



import unittest

import numpy as np
import scipy.stats

from orangecontrib.bioinformatics.widgets.OWDifferentialExpression import f_oneway


class TestFOneWay(unittest.TestCase):
    def test_f_oneway(self):
        g1 = np.array([0.1, -0.1, 0.2, -0.2])
        g2 = g1 + 1
        g3 = g1

        f1, p1 = scipy.stats.f_oneway(g1, g2)
        f, p = f_oneway(g1, g2)
        np.testing.assert_almost_equal([f, p], [f1, p1])

        f, p = f_oneway(np.c_[g1], np.c_[g2], axis=0)
        np.testing.assert_almost_equal([f[0], p[0]], [f1, p1])

        f1, p1 = scipy.stats.f_oneway(g1, g2, g3)
        f, p = f_oneway(g1, g2, g3)
        np.testing.assert_almost_equal([f, p], [f1, p1])

        g1 = np.random.normal(size=(10, 30))
        g2 = np.random.normal(loc=1, size=(10, 20))
        g3 = np.random.normal(loc=2, size=(10, 10))

        f, p = f_oneway(g1, g2, g3, axis=1)
        self.assertEqual(f.shape, (10,))
        self.assertEqual(p.shape, (10,))

        fp1 = [scipy.stats.f_oneway(g1, g2, g3) for g1, g2, g3 in zip(g1, g2, g3)]

        f1 = [f for f, _ in fp1]
        p1 = [p for _, p in fp1]
        np.testing.assert_almost_equal(f1, f)
        np.testing.assert_almost_equal(p1, p)

        f, p = f_oneway(g1.T, g2.T, g3.T, axis=0)
        np.testing.assert_almost_equal(f1, f)
        np.testing.assert_almost_equal(p1, p)



import unittest
from os.path import basename, normpath

from orangecontrib.bioinformatics.ncbi.homologene import HomoloGene


class TestHomoloGene(unittest.TestCase):
    def setUp(self) -> None:
        self.homology = HomoloGene()
        self.assertEqual(basename(normpath(self.homology.file_path)), 'homologene.tab')

    def test_homology(self):
        self.assertEqual(self.homology.find_homolog('920', '9913'), '407098')
        self.assertEqual(self.homology.find_homolog('920', '10090'), '12504')
        self.assertEqual(self.homology.find_homolog('920', '10116'), '24932')
